@@ -128,7 +128,7 @@ int sg_initialize(char msg[1024], int flag)
 
 /* ts A91227 */
 /** Performs global finalization of the SCSI transport adapter and eventually
-    needed operating system facilities. Releases globally aquired resources.
+    needed operating system facilities. Releases globally acquired resources.
     @param flag  unused yet, submit 0
     @return      1 = success, <=0 = failure
 */  
@@ -873,6 +873,9 @@ int sg_issue_command(struct burn_drive *d, struct command *c)
 		ccb->csio.dxfer_len = 0;
 	}
 
+	/* ts B90523 : Record effective transfer length request for debugging*/
+	c->dxfer_len = ccb->csio.dxfer_len;
+
 	start_time = time(NULL);
 	for (i = 0; !done; i++) {
 
@@ -985,6 +988,8 @@ int sg_issue_command(struct burn_drive *d, struct command *c)
 						timeout_ms, i, !!ignore_error);
 		if (d->cancel)
 			done = 1;
+		if (!done)
+			spc_register_retry(c);
 	} while (!done);
 	ret = 1;
 ex:;
@@ -995,7 +1000,7 @@ ex:;
 
 /* ts B00115 */
 /* Return 1 if the given path leads to a regular file or a device that can be
-   seeked, read and eventually written with 2 kB granularity.
+   fseeked, read and eventually written with 2 kB granularity.
 */
 int burn_os_is_2k_seekrw(char *path, int flag)
 {
