@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2009 - 2015 Thomas Schmitt
+ * Copyright (c) 2009 - 2023 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 
@@ -33,7 +33,7 @@
 #define MAX_ISO_FILE_SECTION_SIZE   0xffffffff
 
 /*
- * When a file need to be splitted in several sections, the maximum size
+ * When a file need to be split in several sections, the maximum size
  * of such sections, but the last one. Set to a multiple of BLOCK_SIZE.
  * Default to 4GB - 2048 = 0xFFFFF800
  */
@@ -59,7 +59,7 @@
 #define ISO_DISC_LABEL_SIZE 129
 
 
-/* The maximum lenght of an specs violating ECMA-119 file identifier.
+/* The maximum length of an specs violating ECMA-119 file identifier.
    The theoretical limit is  254 - 34 - 28 (len of SUSP CE entry) = 192
    Currently the practical limit is 254 - 34 - 96 (non-CE RR entries) - 28 (CE)
 */
@@ -88,7 +88,7 @@
 
 
 /* How many warnings to issue about writing Joliet names which cannot be
-   properly represented in UCS-2 and thus had to be defaultet to '_'.
+   properly represented in UCS-2 and thus had to be defaulted to '_'.
 */
 #define ISO_JOLIET_UCS2_WARN_MAX 3
 
@@ -111,7 +111,7 @@ struct iso_write_opts {
 
     unsigned int aaip :1; /* whether to write eventual ACL and EAs */
 
-    /* allways write timestamps in GMT */
+    /* always write timestamps in GMT */
     unsigned int always_gmt :1;
 
     /*
@@ -183,7 +183,7 @@ struct iso_write_opts {
 
     /**
      * Allow all characters to be part of Volume and Volset identifiers on
-     * the Primary Volume Descriptor. This breaks ISO-9660 contraints, but
+     * the Primary Volume Descriptor. This breaks ISO-9660 constraints, but
      * should work on modern systems.
      */
     unsigned int relaxed_vol_atts :1;
@@ -223,7 +223,7 @@ struct iso_write_opts {
     /**
      * Write AAIP as extension according to SUSP 1.10 rather than SUSP 1.12.
      * I.e. without announcing it by an ER field and thus without the need
-     * to preceed the RRIP fields by an ES and to preceed the AA field by ES.
+     * to precede the RRIP fields by an ES and to precede the AA field by ES.
      * This saves bytes and might avoid problems with readers which dislike
      * ER fields other than the ones for RRIP.
      * On the other hand, SUSP 1.12 frowns on such unannounced extensions
@@ -287,7 +287,7 @@ struct iso_write_opts {
     /**
      * The following options set the default values for files and directory
      * permissions, gid and uid. All these take one of three values: 0, 1 or 2.
-     * If 0, the corresponding attribute will be kept as setted in the IsoNode.
+     * If 0, the corresponding attribute will be kept as set in the IsoNode.
      * Unless you have changed it, it corresponds to the value on disc, so it
      * is suitable for backup purposes. If set to 1, the corresponding attrib.
      * will be changed by a default suitable value. Finally, if you set it to
@@ -349,7 +349,7 @@ struct iso_write_opts {
      * or write to an .iso file for future burning or distribution.
      *
      * On the other side, an appendable image is not self contained, it refers
-     * to serveral files that are stored outside the image. Its usage is for
+     * to several files that are stored outside the image. Its usage is for
      * multisession discs, where you add data in a new session, while the
      * previous session data can still be accessed. In those cases, the old
      * data is not written again. Instead, the new image refers to it, and thus
@@ -381,7 +381,7 @@ struct iso_write_opts {
     /**
      * When not NULL, it should point to a buffer of at least 64KiB, where
      * libisofs will write the contents that should be written at the beginning
-     * of a overwriteable media, to grow the image. The growing of an image is
+     * of a overwritable media, to grow the image. The growing of an image is
      * a way, used by first time in growisofs by Andy Polyakov, to allow the
      * appending of new data to non-multisession media, such as DVD+RW, in the
      * same way you append a new session to a multisession disc, i.e., without
@@ -393,7 +393,7 @@ struct iso_write_opts {
      *
      * You should initialize the buffer either with 0s, or with the contents of
      * the first blocks of the image you're growing. In most cases, 0 is good
-     * enought.
+     * enough.
      */
     uint8_t *overwrite;
 
@@ -476,12 +476,19 @@ struct iso_write_opts {
     char *efi_boot_partition;
     int efi_boot_part_flag;
 
-    /* Eventual disk file paths of prepared images which shall be appended
-       after the ISO image and described by partiton table entries in a MBR
+    /* Disk file paths of prepared images which shall be appended
+       after the ISO image and described by partition table entries in a MBR.
+       NULL means unused.
     */
     char *appended_partitions[ISO_MAX_PARTITIONS]; 
     uint8_t appended_part_types[ISO_MAX_PARTITIONS];
     int appended_part_flags[ISO_MAX_PARTITIONS];
+    uint8_t appended_part_type_guids[ISO_MAX_PARTITIONS][16];
+
+    /* Flags in case that appended partitions show up in GPT:
+       bit0= appended_part_type_guids is valid
+    */
+    uint8_t appended_part_gpt_flags[ISO_MAX_PARTITIONS];
 
     /* If 1: With appended partitions: create protective MBR and mark by GPT
      */
@@ -496,6 +503,22 @@ struct iso_write_opts {
      */
     int part_like_isohybrid;
 
+    /* The type to use for the mountable ISO partition if there is any and if
+       the type is not mandatorily determined for particular circumstances like
+       compliant GPT, CHRP, or PReP.
+       -1 = use the default value (e.g. 0xcd, 0x83, 0x17)
+       0x00 to 0xff = value to use if possible 
+    */
+    int iso_mbr_part_type;
+
+    /* iso_write_opts_set_iso_type_guid
+     */
+    uint8_t iso_gpt_type_guid[16];
+    /* bit0= iso_gpt_type_guid is valid
+    */
+    int iso_gpt_flag;
+
+
     /* Eventual name of the non-ISO aspect of the image. E.g. SUN ASCII label.
      */
     char ascii_disc_label[ISO_DISC_LABEL_SIZE];
@@ -513,6 +536,24 @@ struct iso_write_opts {
      */
     int apm_block_size;
 
+    /* User defined GUID for GPT header and base of reproducible partition
+       GUIDs. (Not to be confused with volume "UUID", which is actually a
+       timestamp.)
+       See API call iso_write_opts_set_gpt_guid().
+     */
+    uint8_t gpt_disk_guid[16];
+    int gpt_disk_guid_mode;
+
+    /* Maximum number of CE entries per file */
+    uint32_t max_ce_entries;
+    /* Whether to try dropping AAIP data on too many CE:
+       bit0-3 = Mode:
+                0 = throw ISO_TOO_MANY_CE, without trying to drop anything
+                1 = drop non-isofs fattr
+                2 = drop ACL if dropping non-isofs fattr does not suffice
+    */
+    int max_ce_drop_attr;
+
 };
 
 typedef struct ecma119_image Ecma119Image;
@@ -784,6 +825,11 @@ struct ecma119_image
     */
     IsoFileSrc *sparc_core_src;
 
+    /* Trailing padding of ISO filesystem partition for cylinder alignment */
+    /* Only in effect with Libisofs_part_align_writeR */
+    uint32_t part_align_blocks;
+    uint32_t alignment_end_block;
+
     /* Counted in blocks of 2048 */
     uint32_t appended_part_prepad[ISO_MAX_PARTITIONS];
     uint32_t appended_part_start[ISO_MAX_PARTITIONS];
@@ -844,6 +890,11 @@ struct ecma119_image
     /* Whether the eventual backup GPT is not part of the ISO filesystem */
     int gpt_backup_outside;
 
+    /* The base UUID for the generated GPT UUIDs */
+    uint8_t gpt_uuid_base[16];
+    /* The counter which distinguishes the GPT UUIDs */
+    uint32_t gpt_uuid_counter;
+
     uint32_t efi_boot_part_size;
     IsoFileSrc *efi_boot_part_filesrc; /* Just a pointer. Do not free. */
 
@@ -873,6 +924,15 @@ struct ecma119_image
     uint32_t filesrc_start;
     uint32_t filesrc_blocks;
 
+    /* Number of CE entries in currently processed node */
+    uint32_t curr_ce_entries;
+
+    /* Count of symbolic links and special files which could not be represented
+       in Joliet.
+    */
+    unsigned long joliet_symlinks;
+    unsigned long joliet_specials;
+
 };
 
 #define BP(a,b) [(b) - (a) + 1]
@@ -1012,5 +1072,29 @@ int iso_write_partition_file(Ecma119Image *target, char *path,
 
 void issue_ucs2_warning_summary(size_t failures);
 
+/* Tells whether ivr is a reader from imported_iso in a multi-session
+   add-on situation, and thus to be kept in place.
+*/
+int iso_interval_reader_keep(Ecma119Image *target,
+                             struct iso_interval_reader *ivr,
+                             int flag);
+
+/* @return: ISO_SUCCESS = ok, ISO_SUCCESS + 1 = keep , < 0 = error */
+int iso_interval_reader_start_size(Ecma119Image *t, char *path,
+                                   off_t *start_byte, off_t *byte_count,
+                                   int flag);
+
+/* Obtains start and end number of appended partition range and returns
+   the number of valid entries in the list of appended partitions.
+*/
+int iso_count_appended_partitions(Ecma119Image *target,
+                                  int *first_partition, int *last_partition);
+
+/* Determines the range of valid partition numbers depending on partition
+   table type.
+*/
+void iso_tell_max_part_range(IsoWriteOpts *opts,
+                             int *first_partition, int *last_partition,
+                             int flag);
 
 #endif /*LIBISO_ECMA119_H_*/
