@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2009 - 2011 Thomas Schmitt
+ * Copyright (c) 2009 - 2022 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -28,15 +28,19 @@
  * Create a new IsoFilesystem to deal with local filesystem.
  * 
  * @return
- *     1 sucess, < 0 error
+ *     1 success, < 0 error
  */
 int iso_local_filesystem_new(IsoFilesystem **fs);
 
 
 /* Rank two IsoFileSource of ifs_class by their eventual old image LBAs.
-   Other IsoFileSource classes will be ranked only roughly.
+ * @param cmp_ret  will return the reply value -1, 0, or 1.
+ * @return         1= *cmp_ret is a valid reply
+ *                 0= not both streams are of ifs_class,
+ *                    *cmp_ret is only a rough estimation.
 */
-int iso_ifs_sections_cmp(IsoFileSource *s1, IsoFileSource *s2, int flag);
+int iso_ifs_sections_cmp(IsoFileSource *s1, IsoFileSource *s2, int *cmp_ret,
+                         int flag);
 
 
 /* Create an independent copy of an ifs_class IsoFileSource.
@@ -45,4 +49,14 @@ int iso_ifs_source_clone(IsoFileSource *old_source, IsoFileSource **new_source,
                          int flag);
 
 
+off_t iso_file_source_lseek_capacity(IsoFileSource *src, off_t wanted_size, 
+                                     int flag);
+
+/* Determine whether src is random-access readable and return its capacity.
+*/
+off_t iso_file_source_determine_capacity(IsoFileSource *src, off_t wanted_size,
+                                         int flag);
+
+
+
 #endif /*LIBISO_FSOURCE_H_*/
