@@ -1,7 +1,7 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
  * Copyright (c) 2007 Mario Danic
- * Copyright (c) 2011-2014 Thomas Schmitt
+ * Copyright (c) 2011-2018 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 
@@ -301,14 +301,36 @@ int create_tree(Ecma119Image *t, IsoNode *iso, JolietNode **tree, int pathlen)
         }
         break;
     case LIBISO_SYMLINK:
-    case LIBISO_SPECIAL:
-        {
+        t->joliet_symlinks++;
+        if (t->joliet_symlinks == 1) {
             char *ipath = iso_tree_get_node_path(iso);
+            /* This first ret might indicate the need to abort */
             ret = iso_msg_submit(t->image->id, ISO_FILE_IGNORED, 0,
-                 "Cannot add %s to Joliet tree. %s can only be added to a "
-                 "Rock Ridge tree.", ipath, (iso->type == LIBISO_SYMLINK ?
-                                             "Symlinks" : "Special files"));
+               "Cannot add %s to Joliet tree. Symlinks can only be added to a "
+               "Rock Ridge tree.", ipath);
             free(ipath);
+        } else {
+            if (t->joliet_symlinks == 2)
+                iso_msg_submit(t->image->id, ISO_FILE_IGNORED, 0,
+                         "More symbolic links were omitted from Joliet tree.");
+            ret = 0;
+        }
+        break;
+    case LIBISO_SPECIAL:
+        t->joliet_specials++;
+        if (t->joliet_specials == 1) {
+            char *ipath = iso_tree_get_node_path(iso);
+            /* This first ret might indicate the need to abort */
+            ret = iso_msg_submit(t->image->id, ISO_FILE_IGNORED, 0,
+                       "Cannot add %s to Joliet tree. "
+                       "Special files can only be added to a Rock Ridge tree.",
+                       ipath);
+            free(ipath);
+        } else {
+            if (t->joliet_specials == 2)
+                iso_msg_submit(t->image->id, ISO_FILE_IGNORED, 0,
+                          "More special files were omitted from Joliet tree.");
+            ret = 0;
         }
         break;
     default:
@@ -401,6 +423,20 @@ int joliet_create_mangled_name(uint16_t *dest, uint16_t *src, int digits,
     return ISO_SUCCESS;
 }
 
+/*
+ * From Joliet specs:
+ * "ISO 9660 (Section 7.5.1) states that the sum of the following shall not
+ *  exceed 30:
+ *  - If there is a file name, the length of the file name.
+ *  - If there is a file name extension, the length of the file name extension.
+ *  On Joliet compliant media, however, the sum as calculated above shall not
+ *  exceed 128 [bytes], to allow for longer file identifiers."
+ *
+ * I.e. the dot does not count.
+ *
+ * (We have an option to lift the limit from 64*2 to 103*2, which is the
+ *  maximum to fit into an ISO 9660 directory record.)
+ */
 static
 int mangle_single_dir(Ecma119Image *t, JolietNode *dir)
 {
@@ -455,7 +491,7 @@ int mangle_single_dir(Ecma119Image *t, JolietNode *dir)
         }
 
         /*
-         * A max of 7 characters is good enought, it allows handling up to
+         * A max of 7 characters is good enough, it allows handling up to
          * 9,999,999 files with same name.
          */
          /* Important: joliet_create_mangled_name() relies on digits < 8 */
@@ -481,25 +517,26 @@ int mangle_single_dir(Ecma119Image *t, JolietNode *dir)
                 ext = dot + 1;
 
                 extlen = ucslen(ext);
-                max = maxchar + 1 - extlen - 1 - digits;
+                max = maxchar - extlen - digits;
                 if (max <= 0) {
-                    /* this can happen if extension is too long */
-                    if (extlen + max > 3) {
+                    /*
+                     * This can happen if the extension is too long.
+                     * Reduce its length, to give name at least one
+                     * original character, if it has any.
+                     */
+                    max = (dot > full_name);
+                    extlen = maxchar - max - digits;
+                    if (extlen < 3) {
                         /*
-                         * reduce extension len, to give name an extra char
-                         * note that max is negative or 0
-                         */
-                        extlen = extlen + max - 1;
-                        ext[extlen] = 0;
-                        max = maxchar + 2 - extlen - 1 - digits;
-                    } else {
-                        /*
-                         * error, we don't support extensions < 3
-                         * This can't happen with current limit of digits.
+                         * error, we do not reduce extensions to length < 3
+                         *
+                         * This cannot happen with current limit of digits
+                         * because maxchar is at least 64 and digits at most 7.
                          */
                         ret = ISO_ERROR;
                         goto mangle_cleanup;
                     }
+                    ext[extlen] = 0;
                 }
                 /* ok, reduce name by digits */
                 if (name + max < dot) {
@@ -508,10 +545,10 @@ int mangle_single_dir(Ecma119Image *t, JolietNode *dir)
             } else {
                 /* Directory, or file without extension */
                 if (children[i]->type == JOLIET_DIR) {
-                    max = maxchar + 1 - digits;
+                    max = maxchar - digits;
                     dot = NULL; /* dots have no meaning in dirs */
                 } else {
-                    max = maxchar + 1 - digits;
+                    max = maxchar - digits;
                 }
                 name = full_name;
                 if ((size_t) max < ucslen(name)) {
@@ -1093,7 +1130,7 @@ int write_path_table(Ecma119Image *t, JolietNode **pathlist, int l_type)
     int ret= ISO_SUCCESS;
     uint8_t *zeros = NULL;
 
-    /* 256 is just a convenient size large enought */
+    /* 256 is just a convenient size large enough */
     LIBISO_ALLOC_MEM(buf, uint8_t, 256);
     LIBISO_ALLOC_MEM(zeros, uint8_t, BLOCK_SIZE);
     path_table_size = 0;
