@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2009 - 2011 Thomas Schmitt
+ * Copyright (c) 2009 - 2023 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License version 2 
@@ -30,6 +30,29 @@
 #endif
 
 
+/* Maximum length of a leaf name in the libisofs node tree. This is currently
+   restricted by the implemented maximum length of a Rock Ridge name.
+   This might later become larger and may then be limited to smaller values.
+
+   Rock Ridge specs do not impose an explicit limit on name length.
+   But 255 is also specified by
+     http://pubs.opengroup.org/onlinepubs/009695399/basedefs/limits.h.html
+   which says
+     NAME_MAX >= _XOPEN_NAME_MAX = 255 
+*/
+#define LIBISOFS_NODE_NAME_MAX 255
+
+
+/* Maximum length of a path in the libisofs node tree.
+   Rock Ridge specs do not impose an explicit limit on path length.
+
+     http://pubs.opengroup.org/onlinepubs/009695399/basedefs/limits.h.html
+   says
+     PATH_MAX >= _XOPEN_PATH_MAX = 1024
+*/
+#define LIBISOFS_NODE_PATH_MAX 1024
+
+
 /**
  * The extended information is a way to attach additional information to each
  * IsoNode. External applications may want to use this extension system to
@@ -126,8 +149,15 @@ struct Iso_File
 {
     IsoNode node;
 
+    /* 1 = The node was loaded from an existing ISO image and still refers
+           to its data content there.
+    */
     unsigned int from_old_session : 1;
 
+    /* 1 = The node got attributed a weight by iso_node_set_sort_weight().
+    */
+    unsigned int explicit_weight : 1;
+
     /**
      * It sorts the order in which the file data is written to the CD image.
      * Higher weighting files are written at the beginning of image
@@ -254,7 +284,7 @@ int iso_node_new_symlink(char *name, char *dest, IsoSymlink **link);
  * Create a new special file node. As far as libisofs concerns,
  * an special file is a block device, a character device, a FIFO (named pipe)
  * or a socket. You can choose the specific kind of file you want to add
- * by setting mode propertly (see man 2 stat).
+ * by setting mode properly (see man 2 stat).
  *
  * Note that special files are only written to image when Rock Ridge
  * extensions are enabled. Moreover, a special file is just a directory entry
@@ -286,7 +316,7 @@ int iso_node_new_special(char *name, mode_t mode, dev_t dev,
  * Check if a given name is valid for an iso node.
  *
  * @return
- *     1 if yes, 0 if not
+ *     1 if yes, <0 if not. The value is a specific ISO_* error code.
  */
 int iso_node_is_valid_name(const char *name);
 
@@ -392,6 +422,12 @@ int iso_aa_get_attrs(unsigned char *aa_string, size_t *num_attrs,
 int iso_aa_lookup_attr(unsigned char *aa_string, char *name,
                        size_t *value_length, char **value, int flag);
 
+/**
+ * Delete variables of all namespaces except isofs
+ *
+ * @param flag  bit0= delete ACL, too
+*/
+int iso_node_remove_fattr(IsoNode *node, int flag);
 
 /**
  * Function to identify and manage ZF parameters which do not stem from ZF
@@ -407,9 +443,10 @@ int zisofs_zf_xinfo_func(void *data, int flag);
  * Parameter structure which is to be managed by zisofs_zf_xinfo_func.
  */
 struct zisofs_zf_info {
-    uint32_t uncompressed_size;
+    uint64_t uncompressed_size;
     uint8_t header_size_div4;
     uint8_t block_size_log2;
+    uint8_t zisofs_algo[2];
 };
 
 /**
@@ -457,7 +494,7 @@ int iso_node_cmp_flag(IsoNode *n1, IsoNode *n2, int flag);
 
 
 /**
- * Set the checksum index (typically comming from IsoFileSrc.checksum_index)
+ * Set the checksum index (typically coming from IsoFileSrc.checksum_index)
  * of a regular file node. The index is encoded as xattr "isofs.cx" with
  * four bytes of value.
  */
@@ -482,6 +519,16 @@ int iso_root_get_isofsca(IsoNode *node, uint32_t *start_lba, uint32_t *end_lba,
                          int flag);
 
 
+/**
+ * Record and get truncation parameters as of iso_image_set_truncate_mode() by
+ * "isofs.nt".
+ */
+int iso_root_set_isofsnt(IsoNode *node, uint32_t truncate_mode,
+                         uint32_t truncate_length, int flag);
+int iso_root_get_isofsnt(IsoNode *node, uint32_t *truncate_mode,
+                         uint32_t *truncate_length, int flag);
+
+
 /**
  * Copy the xinfo list from one node to the another.
  */
@@ -513,4 +560,10 @@ int zisofs_zf_xinfo_func(void *data, int flag);
 int zisofs_zf_xinfo_cloner(void *old_data, void **new_data, int flag);
 
 
+/* Performing search for possibly truncated node name.
+ */
+int iso_dir_get_node_trunc(IsoDir *dir, int truncate_length,
+                           const char *name, IsoNode **node);
+
+
 #endif /*LIBISO_NODE_H_*/
