@@ -1,7 +1,7 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
  * Copyright (c) 2007 Mario Danic
- * Copyright (c) 2009 - 2012 Thomas Schmitt
+ * Copyright (c) 2009 - 2022 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -17,7 +17,7 @@
 #include "libisofs.h"
 #include "messages.h"
 #include "joliet.h"
-#include "../version.h"
+#include "node.h"
 
 #include <stdlib.h>
 #include <wchar.h>
@@ -435,6 +435,7 @@ int str2ascii(const char *icharset, const char *input, char **output)
 
         ret_ = malloc(numchars + 1);
         if (ret_ == NULL) {
+            free(wsrc_);
             return ISO_OUT_OF_MEM;
         }
         outbytes = numchars;
@@ -444,7 +445,9 @@ int str2ascii(const char *icharset, const char *input, char **output)
         conv_ret = iso_iconv_open(&conv, "ASCII", "WCHAR_T", 0);
         if (conv_ret <= 0) {
             free(wsrc_);
+            wsrc_ = NULL;
             free(ret_);
+            ret = ret_ = NULL;
         }
     } else if (result != (int) ISO_CHARSET_CONV_ERROR)
         return result;
@@ -593,8 +596,10 @@ int str2ucs(const char *icharset, const char *input, uint16_t **output)
         loop_limit = inbytes + 3;
 
         ret_ = malloc((numchars+1) * sizeof(uint16_t));
-        if (ret_ == NULL)
+        if (ret_ == NULL) {
+            free(wsrc_);
             return ISO_OUT_OF_MEM;
+        }
         outbytes = numchars * sizeof(uint16_t);
         ret = ret_;
 
@@ -602,7 +607,9 @@ int str2ucs(const char *icharset, const char *input, uint16_t **output)
         conv_ret = iso_iconv_open(&conv, "UCS-2BE", "WCHAR_T", 0);
         if (conv_ret <= 0) {
             free(wsrc_);
+            wsrc_ = NULL;
             free(ret_);
+            ret = ret_ = NULL;
         }
     } else if (result != (int) ISO_CHARSET_CONV_ERROR)
         return result;
@@ -723,8 +730,10 @@ int str2utf16be(const char *icharset, const char *input, uint16_t **output)
             loop_limit = inbytes + 3;
 
             ret_ = malloc((2 * numchars+1) * sizeof(uint16_t));
-            if (ret_ == NULL)
+            if (ret_ == NULL) {
+                free(wsrc_);
                 return ISO_OUT_OF_MEM;
+            }
             outbytes = 2 * numchars * sizeof(uint16_t);
             ret = ret_;
 
@@ -794,18 +803,18 @@ int str2utf16be(const char *icharset, const char *input, uint16_t **output)
     return ISO_SUCCESS;
 }
 
-static int valid_d_char(char c)
+int valid_d_char(char c)
 {
     return (c >= '0' && c <= '9') || (c >= 'A' && c <= 'Z') || (c == '_');
 }
 
-static int valid_a_char(char c)
+int valid_a_char(char c)
 {
     return (c >= ' ' && c <= '"') || (c >= '%' && c <= '?') || 
            (c >= 'A' && c <= 'Z') || (c == '_');
 }
 
-static int valid_j_char(uint16_t c)
+int valid_j_char(uint16_t c)
 {
     return cmp_ucsbe(&c, ' ') != -1 && cmp_ucsbe(&c, '*') && cmp_ucsbe(&c, '/')
         && cmp_ucsbe(&c, ':') && cmp_ucsbe(&c, ';') && cmp_ucsbe(&c, '?') 
@@ -1474,6 +1483,14 @@ void iso_lsb(uint8_t *buf, uint32_t num, int bytes)
         buf[i] = (num >> (8 * i)) & 0xff;
 }
 
+void iso_lsb64(uint8_t *buf, uint64_t num)
+{
+    int i;
+
+    for (i = 0; i < 8; ++i)
+        buf[i] = (num >> (8 * i)) & 0xff;
+}
+
 void iso_msb(uint8_t *buf, uint32_t num, int bytes)
 {
     int i;
@@ -1588,6 +1605,22 @@ void iso_datetime_7(unsigned char *buf, time_t t, int always_gmt)
         gmtime_r(&t, &tm);
         tzoffset = 0;
     }
+
+    if (tm.tm_year < 0) {
+        tm.tm_year = 0;
+        tm.tm_mon = 0;
+        tm.tm_mday = 1;
+        tm.tm_hour = 0;
+        tm.tm_min = 0;
+        tm.tm_sec = 0;
+    } else if (tm.tm_year > 255) {
+        tm.tm_year = 255;
+        tm.tm_mon = 11;
+        tm.tm_mday = 31;
+        tm.tm_hour = 23;
+        tm.tm_min = 59;
+        tm.tm_sec = 59;
+    }
     buf[0] = tm.tm_year;
     buf[1] = tm.tm_mon + 1;
     buf[2] = tm.tm_mday;
@@ -1641,12 +1674,18 @@ void iso_datetime_17(unsigned char *buf, time_t t, int always_gmt)
         tzoffset = 0;
     }
 
-    sprintf((char*)&buf[0], "%04d", tm.tm_year + 1900);
-    sprintf((char*)&buf[4], "%02d", tm.tm_mon + 1);
-    sprintf((char*)&buf[6], "%02d", tm.tm_mday);
-    sprintf((char*)&buf[8], "%02d", tm.tm_hour);
-    sprintf((char*)&buf[10], "%02d", tm.tm_min);
-    sprintf((char*)&buf[12], "%02d", MIN(59, tm.tm_sec));
+    if (tm.tm_year <= -1900) {
+        strcpy((char *) buf, "00010101000000");
+    } else if (tm.tm_year >= 8100) {
+        strcpy((char *) buf, "99991231235959");
+    } else {
+        sprintf((char*)&buf[0], "%04d", tm.tm_year + 1900);
+        sprintf((char*)&buf[4], "%02d", tm.tm_mon + 1);
+        sprintf((char*)&buf[6], "%02d", tm.tm_mday);
+        sprintf((char*)&buf[8], "%02d", tm.tm_hour);
+        sprintf((char*)&buf[10], "%02d", tm.tm_min);
+        sprintf((char*)&buf[12], "%02d", MIN(59, tm.tm_sec));
+    }
     memcpy(&buf[14], "00", 2);
     buf[16] = tzoffset;
 
@@ -1794,6 +1833,7 @@ time_t iso_datetime_read_7(const uint8_t *buf)
     tm.tm_hour = buf[3];
     tm.tm_min = buf[4];
     tm.tm_sec = buf[5];
+    tm.tm_isdst = 0;
 
     return timegm(&tm) - ((int8_t)buf[6]) * 60 * 15;
 }
@@ -1810,8 +1850,9 @@ time_t iso_datetime_read_17(const uint8_t *buf)
     sscanf((char*)&buf[12], "%2d", &tm.tm_sec);
     tm.tm_year -= 1900;
     tm.tm_mon -= 1;
+    tm.tm_isdst = 0;
 
-    return timegm(&tm) - ((int8_t)buf[6]) * 60 * 15;
+    return timegm(&tm) - ((int8_t)buf[16]) * 60 * 15;
 }
 
 /**
@@ -1931,7 +1972,7 @@ char *ucs2str(const char *buf, size_t len)
     
     outbytes = (inbytes+1) * MB_LEN_MAX;
     
-    /* ensure enought space */
+    /* ensure enough space */
     out = calloc(outbytes, 1);
     if (out == NULL)
         return NULL;
@@ -1966,16 +2007,25 @@ ex:;
 
 void iso_lib_version(int *major, int *minor, int *micro)
 {
+
+    *major = iso_lib_header_version_major;
+    *minor = iso_lib_header_version_minor;
+    *micro = iso_lib_header_version_micro;
+
+/* No more: values from version.h generated from version.h.in and
+            macro values defined in configure.ac
+
     *major = LIBISOFS_MAJOR_VERSION;
     *minor = LIBISOFS_MINOR_VERSION;
     *micro = LIBISOFS_MICRO_VERSION;
+*/
 }
 
 int iso_lib_is_compatible(int major, int minor, int micro)
 {
     int cmajor, cminor, cmicro;
     
-    /* for now, the rule is that library is compitable if requested
+    /* for now, the rule is that library is compatible if requested
      * version is lower */
     iso_lib_version(&cmajor, &cminor, &cmicro);
 
@@ -2226,6 +2276,7 @@ unexpected_type:;
         goto ex;
     } else if (range_start != ctx_start_lba) {
         ret = ISO_MD5_TAG_MISPLACED;
+        goto ex;
     }
     ret = iso_md5_clone(ctx, &cloned_ctx);
     if (ret < 0)
@@ -2294,7 +2345,7 @@ int iso_clone_mem(char *in, char **out, size_t size)
     if (*out == NULL)
         return ISO_OUT_OF_MEM;
     memcpy(*out, in, size);
-    return 1;
+    return ISO_SUCCESS;
 }
     
 
@@ -2311,7 +2362,7 @@ int iso_clone_mgtd_mem(char *in, char **out, size_t size)
     (Also accepts capital letters.)
     @param text Input like "42", "223062s", "3m" or "-1g"
     @param flag Bitfield for control purposes:
-                bit0= return -1 rathern than 0 on failure
+                bit0= return -1 rather than 0 on failure
                 bit1= if scaled then compute the last byte of the last unit
     @return The derived value
 */
@@ -2343,3 +2394,119 @@ off_t iso_scanf_io_size(char *text, int flag)
         ret += fac - 1;
     return ret;
 }
+
+
+/* Find backward from idx the start byte of a possible UTF-8 character.
+     https://en.wikipedia.org/wiki/UTF-8#Description
+*/
+static
+int find_utf8_start(char *name, int idx, int flag)
+{
+    unsigned char *uname, uch;
+    int i;
+
+    uname= (unsigned char *) name;
+    if ((uname[idx] & 0xc0) != 0x80)
+        return idx;                                /* not an UTF-8 tail byte */
+    for (i = 0; i < 5 && idx - 1 - i >= 0; i++) {                                                                           /* up to deprecated 6-byte codes */
+        uch = uname[idx - 1 - i];
+        if ((uch & 0xe0) == 0xc0 || (uch & 0xf0) == 0xe0 ||
+            (uch & 0xf8) == 0xf0 || (uch & 0xfc) == 0xf8 ||
+            (uch & 0xfe) == 0xfc)
+            return (idx - 1 - i);                  /* UTF-8 start byte found */
+        if ((uch & 0xc0) != 0x80)
+          return idx;                 /* not an UTF-8 tail byte, so no UTF-8 */
+    }
+    return idx;                                      /* no UTF-8 start found */
+}
+
+/* @param flag bit0= do not issue warning message
+*/
+int iso_truncate_rr_name(int truncate_mode, int truncate_length,
+                         char *name, int flag)
+{
+    int neck, goal, ret, l, i;
+    static int hash_size = 32;
+    void *ctx = NULL;
+    char hashval[16];
+
+    l = strlen(name);
+    if (l <= truncate_length)
+        return ISO_SUCCESS;
+    if (truncate_mode == 0)
+        return ISO_RR_NAME_TOO_LONG;
+
+    /* Compute hash */
+    ret = iso_md5_start(&ctx);
+    if (ret < 0)
+        goto ex;
+    ret = iso_md5_compute(ctx, name, l > 4095 ? 4095 : l);
+    if (ret < 0)
+        goto ex;
+    ret = iso_md5_end(&ctx, hashval);
+    if (ret < 0)
+        goto ex;
+
+    if (!(flag & 1))
+        iso_msg_submit(-1, ISO_RR_NAME_TRUNCATED, 0,
+                     "File name had to be truncated and MD5 marked: %s", name);
+
+    /* Avoid to produce incomplete UTF-8 characters */
+    goal = truncate_length - hash_size - 1;
+    neck = find_utf8_start(name, goal, 0);
+    for (; neck < goal; neck++)
+        name[neck] = '_';
+
+    /* Write colon and hash text over end of truncated name */
+    name[goal] = ':';
+    goal++;
+    for (i = 0; goal < truncate_length - 1 && i < hash_size / 2; goal += 2) {
+        sprintf(name + goal, "%2.2x", *((unsigned char *) (hashval + i)));
+        i++;
+    }
+    name[truncate_length] = 0;
+
+    ret = ISO_SUCCESS;
+ex:;
+    if (ctx != NULL)
+        iso_md5_end(&ctx, hashval);
+    return ret;
+}
+
+/* API */
+int iso_truncate_leaf_name(int mode, int length, char *name, int flag)
+{
+    int ret;
+
+    if (mode < 0 || mode > 1)
+        return ISO_WRONG_ARG_VALUE;
+    if (length < 64 || length > LIBISOFS_NODE_NAME_MAX)
+        return ISO_WRONG_ARG_VALUE;
+    ret = iso_truncate_rr_name(mode, length, name, 1);
+    return ret;
+}
+
+/* API */
+/* @param flag bit0= *now contains the time to be set as nowtime override
+               bit1= disable the nowtime override
+   @return 1= *now is not overridden , 2= *now is overridden
+*/
+int iso_nowtime(time_t *now, int flag)
+{
+    static int now_time_overridden = 0;
+    static time_t now_time_override = 0;
+
+    if (flag & 1) {
+        now_time_overridden = 1;
+        now_time_override = *now;
+    }
+    if (flag & 2) {
+        now_time_overridden = 0;
+    }
+    *now = time(NULL);
+    if (!now_time_overridden)
+        return 1;
+    *now = now_time_override;
+    return 2;
+}
+
