@@ -1,7 +1,7 @@
 
 /* xorriso - creates, loads, manipulates and burns ISO 9660 filesystem images.
 
-   Copyright 2007-2014 Thomas Schmitt, <scdbackup@gmx.net>
+   Copyright 2007-2022 Thomas Schmitt, <scdbackup@gmx.net>
 
    Provided under GPL version 2 or later.
 
@@ -384,6 +384,7 @@ int Findjob_new(struct FindjoB **o, char *start_path, int flag)
  m->errmsg[0]= 0;
  m->errn= 0;
  m->match_count= 0;
+ m->depth= 0;
 
  ret= Exprnode_new(&(m->test_tree), m, NULL, "-find", (m->use_shortcuts)<<1);
  if(ret<=0)
@@ -464,7 +465,7 @@ int Findjob_new_node(struct FindjoB *job, struct ExprnodE **fnode,
                      char *origin, int flag)
 /*
  bit0= open new branch 
- bit1= with bit1 : do not register as sub-node of job->cursor
+ bit1= with bit0 : do not register as sub-node of job->cursor
 */
 {
  int ret;
@@ -483,12 +484,14 @@ int Findjob_new_node(struct FindjoB *job, struct ExprnodE **fnode,
        job->errn= -2;
        sprintf(job->errmsg,
                "Program error while parsing -job : sub branch overwrite");
+       Exprnode_destroy(fnode, 0);
        return(0);
      } else
        job->cursor->sub= f;
    }
  } else {
-   f->up= job->cursor->up;
+   if(job->cursor != NULL)
+     f->up= job->cursor->up;
    f->left= job->cursor;
    if(job->cursor!=NULL)
      job->cursor->right= f;
@@ -750,7 +753,7 @@ int Findjob_endif(struct FindjoB *job, int flag)
  }
  if(job->cursor->up==NULL)
    goto improper_range;
- /* test wether parent node is -if */
+ /* test whether parent node is -if */
  job->cursor= job->cursor->up;
  Exprnode_get_branch(job->cursor, &true_branch, 0);
  if(!Exprnode_is_if(job->cursor,0) || true_branch == NULL) {
@@ -892,6 +895,28 @@ int Findjob_set_num_filter(struct FindjoB *o, int test_type,
 }
 
 
+int Findjob_set_size_filter(struct FindjoB *o, int test_type,
+                            off_t num1, int num2, int flag)
+{
+ struct ExprtesT *t;
+ int ret;
+
+ ret= Findjob_default_and(o, 0);
+ if(ret <= 0)
+   return(ret);
+
+ t= o->cursor->test;
+ t->test_type= test_type;
+ t->arg1= calloc(sizeof(off_t), 1);
+ t->arg2= calloc(sizeof(int), 1);
+ if(t->arg1 == NULL || t->arg2 == NULL)
+   return(-1);
+ *((off_t *) t->arg1)= num1;
+ *((int *) t->arg2)= num2;
+ return(1);
+}
+
+
 int Findjob_set_lba_range(struct FindjoB *o, int start_lba, int count,
                           int flag)
 {
@@ -1106,6 +1131,14 @@ int Findjob_get_found_path(struct FindjoB *o, char **path, int flag)
 }
 
 
+int Findjob_get_last_data_file_block(struct FindjoB *o, uint32_t *lba,
+                                     int flag)
+{
+ *lba= o->last_data_file_block;
+ return(1);
+}
+
+
 int Findjob_get_action(struct FindjoB *o, int flag)
 {
  return(o->action);
