@@ -1,7 +1,7 @@
 
 /* xorriso - creates, loads, manipulates and burns ISO 9660 filesystem images.
 
-   Copyright 2007-2017 Thomas Schmitt, <scdbackup@gmx.net>
+   Copyright 2007-2023 Thomas Schmitt, <scdbackup@gmx.net>
 
    Provided under GPL version 2 or later.
 
@@ -143,12 +143,15 @@ int Xorriso_option_paste_in(struct XorrisO *xorriso, char *iso_rr_path,
  }
  bytecount= num;
  sprintf(xorriso->info_text, "-paste_in from %s to %s, byte %.f to %.f",
-         disk_path, iso_rr_path,
+         iso_rr_path, disk_path,
          (double) startbyte, (double) (startbyte+bytecount));
  Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "DEBUG", 0);
 
+ Xorriso_pacifier_reset(xorriso, 0);
  ret= Xorriso_paste_in(xorriso, disk_path, startbyte, bytecount,
                        iso_rr_path, 0); 
+ Xorriso_pacifier_callback(xorriso, "files restored",xorriso->pacifier_count,
+                           xorriso->pacifier_total, "", 1 | 4 | 8 | 32);
  return(ret);
 }
 
@@ -218,8 +221,7 @@ ex:;
  xorriso->allow_graft_points= allow_graft_points_mem;
  Sfile_make_argv("", "", &argc, &argv, 2);
  Xorriso_free_meM(pathspec);
- if(flag & 1)
-   Xorriso_read_lines(xorriso, fp, &linecount, &argc, &argv, 2);
+ Xorriso_read_lines(xorriso, fp, &linecount, &argc, &argv, 2);
 
  if(fp != NULL && fp != stdin)
    fclose(fp);
@@ -303,9 +305,9 @@ int Xorriso_option_print(struct XorrisO *xorriso, char *text, int flag)
    maxl= sizeof(xorriso->mark_text);
  else
    maxl= sizeof(xorriso->result_line);
- if(l >= maxl) {
+ if(l > maxl - 2) {
    sprintf(xorriso->info_text, "Output text too long for -print%s(%d > %d)",
-           mode == 1 ? "_info" : mode == 2 ? "_mark" : "", l, maxl);
+           mode == 1 ? "_info" : mode == 2 ? "_mark" : "", l, maxl - 2);
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "WARNING", 0);
    return(0);
  }
@@ -694,7 +696,7 @@ int Xorriso_option_rollback(struct XorrisO *xorriso, int flag)
  ret= Xorriso_give_up_drive(xorriso, 1|8);
  if(ret<=0)
    goto ex;
- xorriso->image_start_mode&= ~(1<<31); /* reactivate eventual -load address */
+ xorriso->image_start_mode&= ~(1u<<31); /* reactivate eventual -load address */
  ret= Xorriso_option_dev(xorriso, indev, 1|4);
 ex:;
  Xorriso_free_meM(indev);
@@ -831,12 +833,13 @@ int Xorriso_option_session_log(struct XorrisO *xorriso, char *path, int flag)
 int Xorriso_option_setfacl_listi(struct XorrisO *xorriso, char *path, int flag)
 {
  int ret, eaten, line_size;
- size_t buf_size= 0, buf_add= 64 * 1024, l, linecount= 0;
+ size_t buf_size= 0, buf_add, l, linecount= 0, diff_buf_wpt;
  char *line= NULL, *buf= NULL, *wpt, *new_buf, limit_text[80];
  char *file_path= NULL, *uid= NULL, *gid= NULL;
  FILE *fp= NULL;
 
  line_size= SfileadrL * 4;
+ buf_add= line_size * 4;
  Xorriso_alloc_meM(line, char, line_size);
  Xorriso_alloc_meM(file_path, char, SfileadrL);
  Xorriso_alloc_meM(uid, char, 161);
@@ -923,11 +926,13 @@ int Xorriso_option_setfacl_listi(struct XorrisO *xorriso, char *path, int flag)
        Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
        ret= 0; goto ex;
      }
+     diff_buf_wpt= wpt - buf;
      buf_size+= buf_add;
      new_buf= realloc(buf, buf_size);
      if(new_buf == NULL)
        goto out_of_mem;
      buf= new_buf;
+     wpt= buf + diff_buf_wpt;
    }
    memcpy(wpt, line, l);
    *(wpt + l)= '\n';
@@ -989,9 +994,11 @@ int Xorriso_option_setfacli(struct XorrisO *xorriso, char *acl_text,
  ret= Xorriso_normalize_acl_text(xorriso, acl_text,
                                  &access_acl_text, &default_acl_text, 0);
  if(access_acl_text != NULL && default_acl_text != NULL) {
-   sprintf(xorriso->info_text, "Access-ACL :\n%s", access_acl_text);
+   strcpy(xorriso->info_text, "Access-ACL :\n");
+   Xorriso_set_info_text(xorriso, access_acl_text, 2000, 1);
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "DEBUG", 0);
-   sprintf(xorriso->info_text, "Default-ACL :\n%s", default_acl_text);
+   strcpy(xorriso->info_text, "Default-ACL :\n");
+   Xorriso_set_info_text(xorriso, default_acl_text, 2000, 1);
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "DEBUG", 0);
  } else if(access_acl_text == NULL && default_acl_text == NULL) {
    sprintf(xorriso->info_text, "Will delete Access-ACL and Default-ACL");
@@ -1404,12 +1411,31 @@ int Xorriso_option_sleep(struct XorrisO *xorriso, char *duration, int flag)
 /* Commands -speed , -read_speed */
 /* @param flag bit0= -read_speed rather than -speed
 */
-int Xorriso_option_speed(struct XorrisO *xorriso, char *speed, int flag)
+int Xorriso_option_speed(struct XorrisO *xorriso, char *speed_in, int flag)
 {
- int is_cd= 1, unit_found= 0, ret, profile_number, intspeed= 1;
+ int is_cd= 1, unit_found= 0, ret, profile_number, intspeed= 1, for_force= 0;
  double num= -2.0;
- char *cpt, profile_name[80];
+ char *cpt, profile_name[80], *speed;
 
+ speed= speed_in;
+
+ if(strncmp(speed, "soft_corr:", 10) == 0 && (flag & 1)) {
+   sscanf(speed + 10, "%lf", &num);
+   if(num > 1.0e9 || num < 0.0) {
+     sprintf(xorriso->info_text,
+      "-read_speed soft_corr: Value too small or too large (0 to 1e9): '%s'",
+             speed + 10);
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "SORRY", 0);
+     return(0);
+   } 
+   xorriso->read_speed_corr= num;
+   return(1);
+ }
+
+ if(strncmp(speed, "soft_force:", 11) == 0 && (flag & 1)) {
+   for_force= 1;
+   speed+= 11;
+ }
  if(speed[0] == 0 || strcmp(speed, "any") == 0 || strcmp(speed, "max") == 0) {
    intspeed= 0;
  } else if(strcmp(speed, "min") == 0) {
@@ -1421,13 +1447,8 @@ int Xorriso_option_speed(struct XorrisO *xorriso, char *speed, int flag)
    if(num <= 0)
      intspeed= num;
  }
- if(intspeed <= 0) {
-   if(flag & 1)
-     xorriso->read_speed= intspeed;
-   else
-     xorriso->write_speed= intspeed;
-   return(1);
- }
+ if(intspeed <= 0)
+   goto set_speed_and_exit;
 
  for(cpt= speed+strlen(speed)-1; cpt>=speed; cpt--)
    if(isdigit(*cpt) || *cpt=='.')
@@ -1468,17 +1489,24 @@ bd_speed:;
 
  if(num> 2.0e9) {
    sprintf(xorriso->info_text,
-           "-speed: Value too large or not recognizable: '%s'", speed);
+           "%s: Value too large or not recognizable: '%s'",
+           flag & 1 ? "-read_speed" : "-speed", speed);
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "SORRY", 0);
    return(0);
  }
  intspeed= num;
  if(intspeed < num)
    intspeed++;
- if(flag & 1)
-   xorriso->read_speed= intspeed;
- else
+
+set_speed_and_exit:;
+ if(flag & 1) {
+   if(for_force)
+     xorriso->read_speed_force= intspeed;
+   else
+     xorriso->read_speed= intspeed;
+ } else {
    xorriso->write_speed= intspeed;
+ }
  return(1);
 }
 
@@ -1529,18 +1557,18 @@ int Xorriso_option_status_history_max(struct XorrisO *xorriso, int num,
 
 
 /* Option -stdio_sync "on"|"off"|"end"|size */
-int Xorriso_option_stdio_sync(struct XorrisO *xorriso, char *rythm, int flag)
+int Xorriso_option_stdio_sync(struct XorrisO *xorriso, char *rhythm, int flag)
 {
  double num;
 
- if(strcmp(rythm, "default") == 0 || strcmp(rythm, "on") == 0) {
+ if(strcmp(rhythm, "default") == 0 || strcmp(rhythm, "on") == 0) {
    num= 0;
- } else if(strcmp(rythm, "off") == 0) {
+ } else if(strcmp(rhythm, "off") == 0) {
    num= -1;
- } else if(strcmp(rythm, "end") == 0) {
+ } else if(strcmp(rhythm, "end") == 0) {
    num= 1;
  } else {
-   num = Scanf_io_size(rythm, 0) / 2048;
+   num = Scanf_io_size(rhythm, 0) / 2048;
    if(num < 32 || num > 512 * 1024) {
      sprintf(xorriso->info_text,
            "-stdio_sync : Bad size. Use: 64k to 1g, \"on\", \"off\", \"end\"");
@@ -1699,6 +1727,18 @@ int Xorriso_option_toc_of(struct XorrisO *xorriso, char *which, int flag)
 }
 
 
+/* Command -truncate_overwritable */
+int Xorriso_option_truncate_overwritable(struct XorrisO *xorriso,
+                                         char *adr_mode, char *adr_value,
+                                         char *adjust, int flag)
+{
+ int ret;
+
+ ret= Xorriso_truncate_overwritable(xorriso, adr_mode, adr_value, adjust, 0);
+ return(ret);
+}
+
+
 /* Option -uid */
 int Xorriso_option_uid(struct XorrisO *xorriso, char *uid, int flag)
 {
@@ -1989,7 +2029,7 @@ int Xorriso_option_version(struct XorrisO *xorriso, int flag)
  sprintf(xorriso->result_line,
 "ISO 9660 Rock Ridge filesystem manipulator and CD/DVD/BD burn program\n");
  sprintf(xorriso->result_line+strlen(xorriso->result_line),
-"Copyright (C) 2017, Thomas Schmitt <scdbackup@gmx.net>, libburnia project.\n");
+"Copyright (C) 2023, Thomas Schmitt <scdbackup@gmx.net>, libburnia project.\n");
  Xorriso_result(xorriso, 0);
  sprintf(xorriso->result_line,
          "xorriso version   :  %d.%d.%d%s\n",
@@ -2043,6 +2083,8 @@ int Xorriso_option_version(struct XorrisO *xorriso, int flag)
 int Xorriso_option_volid(struct XorrisO *xorriso, char *volid, int flag)
 {
  int warn_shell= 0, warn_ecma= 0, i, ret;
+ char *result= NULL;
+ size_t result_len= 0;
  static char shell_chars[]= {
     "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-+=:.,~@"};
  static char ecma_chars[]= {"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"};
@@ -2064,9 +2106,20 @@ int Xorriso_option_volid(struct XorrisO *xorriso, char *volid, int flag)
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "WARNING", 0);
  }
  if(xorriso->do_joliet && strlen(volid)>16 && !(flag & 1)) {
-   sprintf(xorriso->info_text,
-           "-volid text is too long for Joliet (%d > 16)",(int) strlen(volid));
-   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "WARNING", 0);
+   ret= Xorriso_conv_name_chars(xorriso, volid, &result, &result_len, 2, 0);
+   if(result != NULL)
+     free(result); /* just want the result_len */
+   xorriso->info_text[0]= 0;
+   if(ret <= 0) {
+     sprintf(xorriso->info_text,
+             "Cannot determine length of -volid text in Joliet character set");
+   } else if(result_len > 32) {
+     sprintf(xorriso->info_text,
+             "-volid text is too long for Joliet (%d > 16)",
+             (int) (result_len + 1) / 2);
+   }
+   if(xorriso->info_text[0])
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "WARNING", 0);
  }
  if(warn_ecma && !(flag & 1)) {
    sprintf(xorriso->info_text,
@@ -2211,12 +2264,23 @@ int Xorriso_option_xattr(struct XorrisO *xorriso, char *mode, int flag)
 /* Option -zisofs */
 int Xorriso_option_zisofs(struct XorrisO *xorriso, char *mode, int flag)
 {
- int was_level, was_blocksize, ret, l;
- double num;
- char *cpt, *npt, text[16];
+ int was_level, was_blocksize, was_v2_enabled, was_blocksize_v2;
+ int was_zisofs_susp_z2;
+ uint64_t was_max_total_blocks, was_max_file_blocks, was_block_number_target;
+ int ret, l, i, sticky, set_isofs_params= 0;
+ double num, was_bpt_discard_free_ratio;
+ char *cpt, *npt;
 
  was_level= xorriso->zlib_level;
  was_blocksize= xorriso->zisofs_block_size;
+ was_v2_enabled= xorriso->zisofs_v2_enabled;
+ was_max_total_blocks= xorriso->zisofs_max_total_blocks;
+ was_max_file_blocks= xorriso->zisofs_max_file_blocks;
+ was_blocksize_v2= xorriso->zisofs_v2_block_size;
+ was_block_number_target= xorriso->zisofs_block_number_target;
+ was_bpt_discard_free_ratio= xorriso->zisofs_bpt_discard_free_ratio;
+ was_zisofs_susp_z2= xorriso->zisofs_susp_z2;
+
  npt= cpt= mode;
  for(cpt= mode; npt!=NULL; cpt= npt+1) {
    npt= strchr(cpt,':');
@@ -2240,16 +2304,13 @@ int Xorriso_option_zisofs(struct XorrisO *xorriso, char *mode, int flag)
    } else if(strncmp(cpt, "ziso_used=", 10) == 0 ||
              strncmp(cpt, "osiz_used=", 10) == 0 ||
              strncmp(cpt, "gzip_used=", 10) == 0 ||
-             strncmp(cpt, "gunzip_used=", 12) == 0) {
+             strncmp(cpt, "gunzip_used=", 12) == 0 ||
+             strncmp(cpt, "bpt_ratio_used=", 15) == 0 ||
+             strncmp(cpt, "bpt_used=", 9) == 0) {
      /* (ignored info from -status) */;
 
    } else if(strncmp(cpt, "block_size=", 11)==0) {
-     num= 0.0;
-     if(l > 11 && l < 27) {
-       strncpy(text, cpt + 11, l - 11);
-       text[l - 11]= 0;
-       num= Scanf_io_size(text, 0);
-     }
+     Xorriso__parse_size_param(cpt, 11, l, &num);
      if (num != (1 << 15) && num != (1 << 16) && num != (1 << 17)) {
        sprintf(xorriso->info_text,
                "-zisofs: Unsupported block size (allowed 32k, 64k, 128k)");
@@ -2257,16 +2318,111 @@ int Xorriso_option_zisofs(struct XorrisO *xorriso, char *mode, int flag)
      }
      xorriso->zisofs_block_size= num;
 
-   } else if(strncmp(cpt, "by_magic=", 8)==0) {
-     if(strncmp(cpt + 9, "on", l - 9) == 0)
+   } else if(strncmp(cpt, "by_magic=", 9)==0) {
+     if(strncmp(cpt + 9, "on", l - 9) == 0) {
        xorriso->zisofs_by_magic= 1;
-     else
+     } else if(strncmp(cpt + 9, "v2", l - 9) == 0) {
+       xorriso->zisofs_by_magic= 2;
+     } else if(strncmp(cpt + 9, "off", l - 9) == 0) {
        xorriso->zisofs_by_magic= 0;
+     } else {
+       sprintf(xorriso->info_text,
+         "-zisofs: Unrecognized by_magic mode (allowed: on, off, v2)");
+       goto sorry_ex;
+     }
+
+   } else if(strncmp(cpt, "version_2=", 10) == 0) {
+     if(strncmp(cpt + 10, "off", l - 10) == 0) {
+       xorriso->zisofs_v2_enabled= 0;
+     } else if(strncmp(cpt + 10, "as_needed", l - 10) == 0) {
+       xorriso->zisofs_v2_enabled= 1;
+     } else if(strncmp(cpt + 10, "on", l - 10) == 0) {
+       xorriso->zisofs_v2_enabled= 2;
+     } else {
+       sprintf(xorriso->info_text,
+         "-zisofs: Unrecognized version_2 mode (allowed: on, off, as_needed)");
+       goto sorry_ex;
+     }
+
+   } else if(strncmp(cpt, "max_bpt=", 8) == 0) {
+     Xorriso__parse_size_param(cpt, 8, l, &num);
+     if(num < 1024.0 || num > 128.0 * 1024.0 * 1024.0 * 1024.0) {
+       sprintf(xorriso->info_text,
+         "-zisofs: Unsupported block pointer pool size (allowed: 1k to 128g)");
+       goto sorry_ex;
+     }
+     sticky= 0;
+     if(xorriso->zisofs_max_total_blocks == xorriso->zisofs_max_file_blocks)
+       sticky= 1;
+     xorriso->zisofs_max_total_blocks= num / 8.0;
+     if(xorriso->zisofs_max_total_blocks < xorriso->zisofs_max_file_blocks ||
+        sticky)
+       xorriso->zisofs_max_file_blocks= xorriso->zisofs_max_total_blocks;
+
+   } else if(strncmp(cpt, "max_bpt_f=", 10) == 0) {
+     Xorriso__parse_size_param(cpt, 10, l, &num);
+     if(num < 1024.0 || num > 128.0 * 1024.0 * 1024.0 * 1024.0) {
+       sprintf(xorriso->info_text,
+         "-zisofs: Unsupported block pointer list size (allowed: 1k to 128g)");
+       goto sorry_ex;
+     }
+     xorriso->zisofs_max_file_blocks= num / 8.0;
+     if(xorriso->zisofs_max_file_blocks > xorriso->zisofs_max_total_blocks)
+       xorriso->zisofs_max_total_blocks= xorriso->zisofs_max_file_blocks;
+
+   } else if(strncmp(cpt, "block_size_v2=", 14) == 0) {
+     Xorriso__parse_size_param(cpt, 14, l, &num);
+     for(i= 15 ; i <= 20; i++)
+       if(num == (1 << i))
+     break;
+     if(i > 20) {
+       sprintf(xorriso->info_text,
+        "-zisofs: Unsupported block size (allowed 32k, 64k, 128k, ... 1024k)");
+       goto sorry_ex;
+     }
+     xorriso->zisofs_v2_block_size= num;
+
+   } else if(strncmp(cpt, "bpt_target=", 11) == 0) {
+     Xorriso__parse_size_param(cpt, 11, l, &num);
+     xorriso->zisofs_block_number_target= num;
+
+   } else if(strncmp(cpt, "bpt_free_ratio=", 15) == 0) {
+     Xorriso__parse_size_param(cpt, 15, l, &num);
+     /* 0 means to libisofs "do not change" */
+     if(num == 0.0)
+       num= -1.0;
+     if(num != -1.0 && (num <= 0.0 || num > 1.0)) {
+       sprintf(xorriso->info_text,
+  "-zisofs: Unsupported free blockpointer ratio (allowed -1 or 0.0 to 1.0)");
+       goto sorry_ex;
+     }
+     xorriso->zisofs_bpt_discard_free_ratio = num;
+
+   } else if(strncmp(cpt, "susp_z2=", 8) == 0) {
+     if(strncmp(cpt + 8, "off", l - 8) == 0) {
+       xorriso->zisofs_susp_z2= 0;
+     } else if(strncmp(cpt + 8, "on", l - 8) == 0) {
+       xorriso->zisofs_susp_z2= 1;
+     } else {
+       sprintf(xorriso->info_text,
+         "-zisofs: Unrecognized susp_z2 mode (allowed: on, off)");
+       goto sorry_ex;
+     }
+     Xorriso_set_zisofs_params(xorriso, 2);
 
    } else if(strncmp(cpt, "default", l)==0) {
      xorriso->zlib_level= xorriso->zlib_level_default;
      xorriso->zisofs_block_size= xorriso->zisofs_block_size_default;
      xorriso->zisofs_by_magic= 0;
+     xorriso->zisofs_v2_enabled= 0;
+     xorriso->zisofs_max_total_blocks= xorriso->zisofs_max_total_blocks_default;
+     xorriso->zisofs_max_file_blocks= xorriso->zisofs_max_file_blocks_default;
+     xorriso->zisofs_v2_block_size= xorriso->zisofs_v2_block_size_default;
+     xorriso->zisofs_block_number_target= -1;
+     xorriso->zisofs_bpt_discard_free_ratio=
+         xorriso->zisofs_bpt_discard_free_ratio_default;
+     xorriso->zisofs_susp_z2= xorriso->zisofs_susp_z2_default;
+     Xorriso_set_zisofs_params(xorriso, 2);
 
    } else {
 unknown_mode:;
@@ -2278,10 +2434,24 @@ sorry_ex:
      Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
      xorriso->zlib_level= was_level;
      xorriso->zisofs_block_size= was_blocksize;
+     xorriso->zisofs_v2_enabled= was_v2_enabled;
+     xorriso->zisofs_max_total_blocks= was_max_total_blocks;
+     xorriso->zisofs_max_file_blocks= was_max_file_blocks;
+     xorriso->zisofs_v2_block_size= was_blocksize_v2;
+     xorriso->zisofs_block_number_target= was_block_number_target;
+     xorriso->zisofs_bpt_discard_free_ratio= was_bpt_discard_free_ratio;
+     xorriso->zisofs_susp_z2 = was_zisofs_susp_z2;
+     Xorriso_set_zisofs_params(xorriso, 2);
      return(0);
    }
+
+   if(strncmp(cpt, "by_magic=", 9) != 0 &&
+      strncmp(cpt, "susp_z2=", 8) != 0)
+     set_isofs_params= 1;
  }
- ret= Xorriso_set_zisofs_params(xorriso, 0);
+ ret= 1;
+ if(set_isofs_params)
+   ret= Xorriso_set_zisofs_params(xorriso, 1);
  return(ret);
 }
 
