@@ -2,7 +2,7 @@
 
 /* xorriso - creates, loads, manipulates and burns ISO 9660 filesystem images.
 
-   Copyright 2007-2015 Thomas Schmitt, <scdbackup@gmx.net>
+   Copyright 2007-2022 Thomas Schmitt, <scdbackup@gmx.net>
 
    Provided under GPL version 2 or later.
 
@@ -41,6 +41,7 @@
 #include "iso_manip.h"
 #include "read_run.h"
 #include "sort_cmp.h"
+#include "disk_ops.h"
 
 
 int Xorriso__read_pacifier(IsoImage *image, IsoFileSource *filesource)
@@ -533,6 +534,278 @@ ex:;
 }
 
 
+/* If defined the position accounting will be done by lseek() and used to
+ * verify the position accounting in struct Xorriso_sparse_statE.
+ * # def ine Xorriso_check_sparsE yes
+ */
+
+struct Xorriso_sparse_statE {
+ int use_lseek;
+ off_t cur_pos;
+ off_t after_last_written;
+ int warnings;
+};
+
+
+#ifdef Xorriso_check_sparsE
+
+static int Xorriso_sparse_warn(struct XorrisO *xorriso,
+                               struct Xorriso_sparse_statE *sparse_state,
+                               int occasion, char *msg, int flag)
+{
+ if(sparse_state->warnings & (1 << occasion))
+   return(1);
+ sparse_state->warnings|= 1 << occasion;
+ Xorriso_msgs_submit(xorriso, 0, msg, 0, "SORRY", 0);
+ return(1);
+}
+
+#endif /* Xorriso_check_sparsE */
+
+
+static int Xorriso_sparse_init(struct XorrisO *xorriso,
+                               struct Xorriso_sparse_statE **sparse_state,
+                               int write_fd, int flag)
+{
+ struct Xorriso_sparse_statE *o= NULL;
+ off_t cur_pos;
+ struct stat stbuf;
+ int ret;
+
+ *sparse_state= NULL;
+
+ /* Check whether sparse writing is disabled */
+ if(xorriso->sparse_min_gap <= 0)
+   {ret= 0; goto ex;}
+
+ /* Analyze write_fd */
+ ret= fstat(write_fd, &stbuf);
+ if(ret == -1)
+   {ret= 0; goto ex;}
+ if(!S_ISREG(stbuf.st_mode))
+   {ret= 0; goto ex;}
+ cur_pos= lseek(write_fd, (off_t) 0, SEEK_CUR);
+ if(cur_pos < stbuf.st_size)
+   {ret= 0; goto ex;}
+
+ Xorriso_alloc_meM(o, struct Xorriso_sparse_statE, 1);
+
+ /* Initialize sparse_state */
+ o->use_lseek= 1;
+ o->cur_pos= o->after_last_written= cur_pos;
+ o->warnings= 0;
+
+ ret= 1;
+ex:;
+ if(ret >= 1)
+   *sparse_state= o;
+ else
+   Xorriso_free_meM(o);
+ return(ret);
+}
+
+
+static int Xorriso_sparse_zeroize(struct XorrisO *xorriso,
+                                  struct Xorriso_sparse_statE *sparse_state,
+                                  int write_fd, off_t start, off_t count,
+                                  int flag)
+{
+ int ret, buf_size= 32 * 1024, buf_fill, wret;
+ off_t todo, seek_ret;
+ char *buf= NULL;
+
+ if(count <= 0)
+   {ret= 2; goto ex;}
+ Xorriso_alloc_meM(buf, char, buf_size);
+
+ seek_ret= lseek(write_fd, start, SEEK_SET);
+ if(seek_ret == -1)
+   {ret= -1; goto ex;}
+ sparse_state->cur_pos= seek_ret;
+ for(todo= count; todo > 0; ) {
+   if(buf_size < todo)
+     buf_fill= buf_size;
+   else
+     buf_fill= todo;
+   wret= write(write_fd, buf, buf_fill);
+   if(wret <= 0)
+     {ret= wret; goto ex;}
+   todo-= wret;
+   sparse_state->cur_pos+= wret;
+ }
+ ret= 1;
+ex:;
+ Xorriso_free_meM(buf);
+ return(ret);
+}
+
+
+/* @param flag bit0= this is the last buffer of the stream
+*/
+static int Xorriso_sparse_write(struct XorrisO *xorriso,
+                                struct Xorriso_sparse_statE *sparse_state,
+                                int write_fd, char *buf, int count,
+                                int flag)
+{
+ int wret, i, ret;
+ off_t cur_pos= -1, seek_ret;
+ static char zero[32]= {0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,
+                        0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};
+
+ if(sparse_state == NULL)
+   goto do_write;
+ if(!sparse_state->use_lseek)
+   goto do_write;
+ if(flag & 1)
+   goto do_write;
+
+#ifdef Xorriso_check_sparsE
+
+ cur_pos= lseek(write_fd, (off_t) 0, SEEK_CUR);
+ if(cur_pos == -1)
+   goto do_write;
+ if(cur_pos != sparse_state->cur_pos) {
+   Xorriso_sparse_warn(xorriso, sparse_state, 0,
+                       "cur_pos deviation in Xorriso_sparse_write:intro", 0);
+   sparse_state->cur_pos= cur_pos;
+ }
+
+#else
+  
+ cur_pos= sparse_state->cur_pos;
+
+#endif
+
+ /* Check for all zeros */
+ if(count % 32)
+   goto do_write;
+ for(i= 0; i < count; i+= 32)
+   if(memcmp(buf + i, zero, 32))
+ break;
+ if(i < count)
+   goto do_write;
+
+ /* Omit write() until next non-zero buffer or end of writing */
+
+#ifdef Xorriso_check_sparsE
+
+ /* Only for debugging: Do real lseek() instead of write() */
+ seek_ret= lseek(write_fd, cur_pos + count, SEEK_SET);
+ if(seek_ret == -1)
+   return(-1);
+
+#endif
+
+ sparse_state->cur_pos= cur_pos + count;
+ return(count);
+
+do_write:
+ if(sparse_state != NULL) {
+   /* Check whether the gap since after_last_written is too small.
+      If so: fill the whole gap by writing zeros.
+   */
+   if(sparse_state->after_last_written < cur_pos) {
+     if(xorriso->sparse_min_gap > cur_pos - sparse_state->after_last_written) {
+       ret= Xorriso_sparse_zeroize(xorriso, sparse_state, write_fd,
+                                sparse_state->after_last_written,
+                                cur_pos - sparse_state->after_last_written, 0);
+       if(ret < 0)
+         return(ret);
+       if(ret == 0) {
+         seek_ret= lseek(write_fd, cur_pos, SEEK_SET);
+         if(seek_ret == -1)
+           return(-1);
+         sparse_state->cur_pos= seek_ret;
+       }
+     }
+   }
+ }
+
+
+ if(sparse_state != NULL) {
+
+#ifdef Xorriso_check_sparsE
+
+   cur_pos= lseek(write_fd, (off_t) 0, SEEK_CUR);
+   if(cur_pos != sparse_state->cur_pos) {
+     Xorriso_sparse_warn(xorriso, sparse_state, 1,
+                      "cur_pos deviation in Xorriso_sparse_write:do_write", 0);
+     sparse_state->cur_pos= cur_pos;
+   }
+
+#else
+
+   /* lseek() has been delayed until now */
+   if(sparse_state->after_last_written != sparse_state->cur_pos) {
+     seek_ret= lseek(write_fd, sparse_state->cur_pos, SEEK_SET);
+     if(seek_ret == -1)
+       return(-1);
+   }
+
+#endif /* ! Xorriso_check_sparsE */
+
+ }
+ wret= write(write_fd, buf, count);
+ if(sparse_state != NULL && wret > 0 && cur_pos >= 0)
+   sparse_state->cur_pos= sparse_state->after_last_written= cur_pos + wret;
+ return(wret);
+}
+
+
+static int Xorriso_sparse_finish(struct XorrisO *xorriso,
+                                 struct Xorriso_sparse_statE **sparse_state,
+                                 int write_fd, int flag)
+{
+ int ret;
+ off_t cur_pos;
+ struct Xorriso_sparse_statE *o;
+
+ if(sparse_state == NULL)
+   return(0);
+ o= *sparse_state;
+ if(o == NULL)
+   return(1);
+ if(write_fd == -1)
+   {ret= 1; goto ex;}
+
+#ifdef Xorriso_check_sparsE
+
+ cur_pos= lseek(write_fd, (off_t) 0, SEEK_CUR);
+ if(cur_pos == -1)
+   {ret= -1; goto ex;}
+ if(cur_pos != o->cur_pos) {
+   Xorriso_sparse_warn(xorriso, o, 2,
+                       "cur_pos deviation in Xorriso_sparse_finish", 0);
+   o->cur_pos= cur_pos;
+ }
+
+#else
+  
+ cur_pos= o->cur_pos;
+
+#endif /* ! Xorriso_check_sparsE */
+
+ if(o->after_last_written < cur_pos) {
+   /* Check whether the gap since after_last_written is too small.
+      If so: fill the whole gap by writing zeros, else: write a last zero byte.
+   */
+   if(xorriso->sparse_min_gap > cur_pos - o->after_last_written)
+     ret= Xorriso_sparse_zeroize(xorriso, o, write_fd, o->after_last_written,
+                                 cur_pos - o->after_last_written, 0);
+   else
+     ret= Xorriso_sparse_zeroize(xorriso, o, write_fd, cur_pos - 1, (off_t) 1,
+                                 0);
+   if(ret <= 0)
+     goto ex;
+ }
+ ret= 1;
+ex:;
+ Xorriso_free_meM(o);
+ *sparse_state= NULL;
+ return(ret);
+}
+
+
 /* @param flag bit0= Minimal transfer: access permissions only
                bit1= *_offset and bytes are valid for writing to regular file
                bit2= This is not a parameter. Do not report if ignored
@@ -552,7 +825,7 @@ int Xorriso_tree_restore_node(struct XorrisO *xorriso, IsoNode *node,
  int ret= 0, write_fd= -1, wanted, wret, open_flags, l_errno= 0;
  int target_deleted= 0, buf_size= 32 * 1024;
  char *what= "[unknown filetype]";
- char *buf= NULL, type_text[5], *temp_path= NULL, *buf_pt;
+ char *buf= NULL, type_text[5], *temp_path= NULL, *buf_pt, *reason;
  char *link_target, *open_path_pt= NULL;
  off_t todo= 0, size, seek_ret, last_p_count= 0, already_done, read_count= 0;
  void *data_stream= NULL;
@@ -564,11 +837,17 @@ int Xorriso_tree_restore_node(struct XorrisO *xorriso, IsoNode *node,
  IsoBoot *bootcat;
  uint32_t lba;
  char *catcontent = NULL;
- off_t catsize;
+ off_t catsize, iso_node_size, wanted_size, cap;
+ char disk_md5[16], iso_md5[16];
+ void *ctx= NULL;
+ int use_md5= 0, i, sparse_ret= 3;
+ struct Xorriso_sparse_statE *sparse_state= NULL;
 
  Xorriso_alloc_meM(buf, char, buf_size);
  Xorriso_alloc_meM(temp_path, char, SfileadrL);
 
+ if(!(flag & 2))
+   img_offset= bytes= 0;
  if(LIBISO_ISDIR(node)) {
    what= "directory";
    ret= mkdir(disk_path, 0777);
@@ -583,19 +862,55 @@ int Xorriso_tree_restore_node(struct XorrisO *xorriso, IsoNode *node,
                                 1);
      if(ret<=0)
        goto ex;
+     if((xorriso->do_md5 & 65) == 65 && !(flag & 2)) {
+       ret= Xorriso_is_plain_image_file(xorriso, (void *) node, img_path, 0);
+       if(ret > 0) {
+         ret= Xorriso_get_md5(xorriso, (void *) node, img_path, iso_md5, 1);
+         if(ret > 0)
+           ret= Xorriso_md5_start(xorriso, &ctx, 0);
+         if(ret > 0) {
+           use_md5= 1;
+         } else if(xorriso->do_md5 & 128) {
+           sprintf(xorriso->info_text,
+                   "Cannot obtain any recorded MD5 of file ");
+           Text_shellsafe(img_path, xorriso->info_text, 1);
+           Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "SORRY", 0);
+           ret= Xorriso_eval_problem_status(xorriso, 0, 1 | 2);
+           if(ret < 0)
+             {ret= 0; goto ex;}
+         }
+       }
+     }
    }
    open_path_pt= disk_path;
    ret= stat(open_path_pt, &stbuf);
    if(ret == -1 && errno == EACCES && (flag & 128))
      {ret= 4; goto ex;}
    if(flag&2) {
-     if(ret!=-1 && !S_ISREG(stbuf.st_mode)) {
-       sprintf(xorriso->info_text,
-      "Restore offset demanded. But filesystem path leads to non-data file ");
-       Text_shellsafe(disk_path, xorriso->info_text, 1);
-       Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, errno, "FAILURE",0);
-       l_errno= 0;
-       goto cannot_restore;
+     if(ret != -1) {
+       wanted_size= disk_offset + bytes;
+       iso_node_size= iso_file_get_size((IsoFile *) node);
+       if(wanted_size > disk_offset + iso_node_size)
+         wanted_size= disk_offset + iso_node_size;
+       cap= wanted_size;
+       ret= Xorriso_determine_capacity(xorriso, open_path_pt, &cap, &reason,
+                                       3);
+       if(ret <= 0 || (cap <= 0 && !S_ISREG(stbuf.st_mode))) {
+         if(ret > 0)
+           reason= "has addressable range 0";
+         sprintf(xorriso->info_text,
+                 "Restore offset demanded. But target file (%s) ",
+                 Ftypetxt(stbuf.st_mode, 0));
+         Text_shellsafe(disk_path, xorriso->info_text, 1);
+         strcat(xorriso->info_text, reason);
+         Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE",0);            l_errno= 0;
+         goto cannot_restore;
+       }
+       if(cap < wanted_size && !S_ISREG(stbuf.st_mode)) {
+
+         /* >>> non-regular file might be too small to take the interval */;
+
+       }
      }
    } else {
      /* If source and target are the same disk file then do not copy content */
@@ -655,6 +970,9 @@ int Xorriso_tree_restore_node(struct XorrisO *xorriso, IsoNode *node,
        goto cannot_restore;
      }
    }
+   if(ISO_NODE_IS_FILE(node))
+     Xorriso_sparse_init(xorriso, &sparse_state, write_fd, 0);
+
    while(todo>0) {
      wanted= buf_size;
      if(wanted>todo)
@@ -696,6 +1014,8 @@ int Xorriso_tree_restore_node(struct XorrisO *xorriso, IsoNode *node,
    break;
      }
      read_count+= ret;
+     if(use_md5)
+       Xorriso_md5_compute(xorriso, ctx, buf_pt, ret, 0);
 
      if(img_offset > read_count - ret) {
        /* skip the desired amount of bytes */
@@ -705,7 +1025,11 @@ int Xorriso_tree_restore_node(struct XorrisO *xorriso, IsoNode *node,
        ret= read_count - img_offset;
      }
 
-     wret= write(write_fd, buf_pt, ret);
+     if(sparse_state == NULL)
+       wret= write(write_fd, buf_pt, ret);
+     else
+       wret= Xorriso_sparse_write(xorriso, sparse_state, write_fd, buf_pt, ret,
+                                  0);
      if(wret>=0) {
        todo-= wret;
        xorriso->pacifier_byte_count+= wret;
@@ -724,9 +1048,35 @@ int Xorriso_tree_restore_node(struct XorrisO *xorriso, IsoNode *node,
        Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, errno, "FAILURE",0);
    break;
      }
-   }  
-   if(write_fd != -1)
+   }
+  
+   if(use_md5) {
+     ret= Xorriso_md5_end(xorriso, &ctx, disk_md5, 0);
+     if(ret <= 0) {
+       sprintf(xorriso->info_text,
+        "Internal problem with obtaining computed MD5 for extracted data of ");
+       goto bad_md5;
+     } else {
+       for(i= 0; i < 16; i++)
+         if(iso_md5[i] != disk_md5[i])
+       break;
+       if(i < 16) {
+         sprintf(xorriso->info_text,
+                 "MD5 of extracted data does not match recorded MD5 of file ");
+bad_md5:;
+         Text_shellsafe(img_path, xorriso->info_text, 1);
+         Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "SORRY", 0);
+         ret= Xorriso_eval_problem_status(xorriso, 0, 1 | 2);
+         if(ret < 0)
+           {ret= 0; goto ex;}
+       }
+     }
+   }
+
+   if(write_fd != -1) {
+     sparse_ret= Xorriso_sparse_finish(xorriso, &sparse_state, write_fd, 0);
      close(write_fd);
+   }
    write_fd= -1;
    if(todo > 0 && xorriso->extract_error_mode == 2) {
      unlink(open_path_pt);
@@ -749,6 +1099,12 @@ int Xorriso_tree_restore_node(struct XorrisO *xorriso, IsoNode *node,
    }
    ret= -(todo > 0);
    l_errno= 0;
+   if(sparse_ret <= 0) {
+     strcpy(xorriso->info_text, "Could not finalize sparse extraction of ");
+     Text_shellsafe(disk_path, xorriso->info_text, 1 | 2);
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text,
+                         sparse_ret < 0 ? errno : 0, "FAILURE", 0);
+   }
 
  } else if(LIBISO_ISLNK(node)) {
    what= "symbolic link";
@@ -847,6 +1203,10 @@ ex:;
    free(catcontent);
  if(data_stream!=NULL)
    Xorriso_iso_file_close(xorriso, &data_stream, 0);
+ if(ctx != NULL)
+   Xorriso_md5_end(xorriso, &ctx, disk_md5, 0);
+ if(sparse_state != NULL)
+   Xorriso_sparse_finish(xorriso, &sparse_state, -1, 0);
  Xorriso_process_msg_queues(xorriso,0);
  return(ret);
 }
@@ -1036,6 +1396,7 @@ int Xorriso_register_node_target(struct XorrisO *xorriso, int node_idx,
                bit6= this is a copy action: do not fake times and ownership
                bit7= return 4 if restore fails from denied permission
                      do not issue error message
+              bit10= do not restore properties
    @return <=0 = error , 1 = added leaf file object , 2 = added directory ,
              3= regularly not installed (disallowed device, UNIX domain socket)
              4 = with bit7: permission to restore was denied
@@ -1046,7 +1407,7 @@ int Xorriso_restore_disk_object(struct XorrisO *xorriso,
                                 off_t offset, off_t bytes, int flag)
 {
  int ret, i, split_count= 0, partno, total_parts, leaf_is_split= 0;
- int record_hl_path= 0, node_idx, cannot_register= 0;
+ int record_hl_path= 0, node_idx, cannot_register= 0, no_props;
  off_t total_bytes;
  char *part_name, *part_path= NULL, *img_path_pt= NULL;
  IsoImage *volume;
@@ -1056,13 +1417,15 @@ int Xorriso_restore_disk_object(struct XorrisO *xorriso,
 
  Xorriso_alloc_meM(part_path, char, SfileadrL);
 
+ no_props= !!(flag & 1024);
+
  ret= Xorriso_get_volume(xorriso, &volume, 0);
  if(ret<=0)
    goto ex;
 
  if(LIBISO_ISDIR(node) && xorriso->do_concat_split)
    leaf_is_split= Xorriso_identify_split(xorriso, img_path, node,
-                                      &split_parts, &split_count, &stbuf, 1|2);
+                                    &split_parts, &split_count, &stbuf, 1 | 4);
  if(leaf_is_split) {
    /* map all files in directory img_path into regular file disk_path */
 
@@ -1090,7 +1453,7 @@ int Xorriso_restore_disk_object(struct XorrisO *xorriso,
      if(ret == 4)
        goto ex;
    }
-   if(first_part_node!=NULL)
+   if(first_part_node != NULL && !no_props)
      Xorriso_restore_properties(xorriso, disk_path, first_part_node,
                                 !!(flag&64));
    goto went_well;
@@ -1132,10 +1495,11 @@ int Xorriso_restore_disk_object(struct XorrisO *xorriso,
 
  ret= Xorriso_tree_restore_node(xorriso, node, img_path_pt, (off_t) 0,
                                 disk_path, offset, bytes,
-     (flag&(4 | 8 | 128)) | (!!(flag&64)) | ((flag&1)<<1) | ( 16 * !(flag&2)));
+     (flag&(4 | 8 | 128)) | (!!(flag&64)) | ((flag&1)<<1) | ( 16 * !(flag&2))
+     | (no_props << 3));
  if(ret == 4)
    goto ex;
- if(ret>0 && (flag&8))
+ if(ret > 0 && (flag & 8) && !no_props)
    ret= Xorriso_restore_properties(xorriso, disk_path, node, 2 | !!(flag&64));
  if(ret<=0) {
 restoring_failed:;
@@ -1249,7 +1613,6 @@ ex:;
 /* @param flag bit0= recursion is active
                bit1= do not report restored files
                bit6= this is a copy action: do not fake times and ownership
-               bit8= only register non-directory nodes in xorriso->node_array
                bit7+8=
                      0= direct operation
                      1= create only directories,
@@ -1389,9 +1752,14 @@ much_too_long:;
      {ret= 0; goto was_problem;}
    if(Xorriso_much_too_long(xorriso, len_ip + strlen(leaf_name)+1, 0)<=0)
      {ret= 0; goto was_problem;}
-   /* name is a pointer into img_path */
+   /* name and disk_name are pointers into img_path and disk_path */
    strcpy(name, leaf_name);
    strcpy(disk_name,  leaf_name);
+   ret= Xorriso_path_is_excluded(xorriso, disk_path, 2);
+   if(ret < 0)
+     goto was_problem;
+   if(ret > 0)
+ continue;
 
    stbuf_src= srcpt;
    ret= Xorriso_fake_stbuf(xorriso, img_path, &stbuf, &node, 1);
@@ -1437,7 +1805,7 @@ much_too_long:;
      source_is_dir= 1;
    source_is_split= 0;
    if(source_is_dir)
-     source_is_split= Xorriso_is_split(xorriso, img_path, node, 1|2);
+     source_is_split= Xorriso_is_split(xorriso, img_path, node, 1 | 2 | 4);
    if(source_is_split)
      do_not_dive= 1;
 
@@ -1469,6 +1837,8 @@ much_too_long:;
      } else {
        ret= Xorriso_restore_disk_object(xorriso, img_path, node, disk_path,
                                         (off_t) 0, (off_t) 0, hflag);
+       if(ret == 3) /* intentionally not restored */
+         do_not_dive= 1;
      }
      if(ret<=0)
        goto was_problem;
@@ -1509,7 +1879,7 @@ ex:
  if(link_target!=NULL)
    free(link_target);
  Xorriso_findi_iter(xorriso, dir, &mem, &iter, &node_array, &node_count,
-                    &node_idx, &node, (1<<31));
+                    &node_idx, &node, (1u<<31));
 
  Xorriso_process_msg_queues(xorriso,0);
 
@@ -1543,6 +1913,7 @@ ex:
                      3= count nodes in xorriso->node_counter,
                         create no directory
                bit9= with operation mode 1 do net register prefixes
+              bit10= with bit3: do not restore properties to leaf file
    @return <=0 = error , 1 = added leaf file object , 2 = added directory ,
                          3 = rejected 
 */
@@ -1555,7 +1926,7 @@ int Xorriso_restore(struct XorrisO *xorriso,
  IsoNode *node= NULL;
  int done= 0, is_dir= 0, ret, source_is_dir, stbuf_ret, hret;
  int dir_create= 0, node_count= 0, node_register= 0, path_size;
- int leaf_is_split= 0, source_is_split= 0, new_dir_made= 0;
+ int leaf_is_split= 0, source_is_split= 0, new_dir_made= 0, no_props;
  struct stat stbuf;
  struct PermiteM *perm_stack_mem;
 
@@ -1569,6 +1940,7 @@ int Xorriso_restore(struct XorrisO *xorriso,
  break; case 2: node_register= 1;
  break; case 3: node_count= 1;
  }
+ no_props= (!!(flag & 8)) * (flag & 1024);
 
  if(dir_create && !(flag & (1 << 9))) {
    ret= Xorriso_lst_append_binary(&(xorriso->node_disk_prefixes),
@@ -1581,11 +1953,18 @@ int Xorriso_restore(struct XorrisO *xorriso,
      goto ex; 
  }
 
- ret= Xorriso_path_is_excluded(xorriso, disk_path, !(flag&4));
- if(ret<0)
-   goto ex;
- if(ret>0)
+ ret= Xorriso_path_is_excluded(xorriso, disk_path, 4 | 2 | !(flag & 4));
+ if(ret < 0)
+   {ret= 0; goto ex;}
+ if(ret > 0) {
+   if(!(flag & 4)) {
+     strcpy(xorriso->info_text,
+            "Excluded from restoring by -not_path or -not_leaf :");
+     Text_shellsafe(disk_path, xorriso->info_text, 1);
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "SORRY", 0);
+   }
    {ret= 3*!!(flag&16); goto ex;}
+ }
 
  ret= Xorriso_get_volume(xorriso, &volume, 0);
  if(ret<=0)
@@ -1627,7 +2006,7 @@ int Xorriso_restore(struct XorrisO *xorriso,
      ret= 0; goto ex;
    }
    if(is_dir && xorriso->do_concat_split)
-     leaf_is_split= Xorriso_is_split(xorriso, img_path, node, 1|2);
+     leaf_is_split= Xorriso_is_split(xorriso, img_path, node, 1 | 2 | 4);
  }
  for(npt= apt; !done; apt= npt+1) {
    npt= strchr(apt, '/');
@@ -1650,7 +2029,7 @@ int Xorriso_restore(struct XorrisO *xorriso,
    if((flag&8) && done) {
 
      /* ??? move down from Xorriso_paste_in() :
-            check whether target does not exist or both are regular */;
+            check whether target does not exist or both are suitable */;
 
    } else if(source_is_dir || !(dir_create || node_count || node_register)) {
      ret= Xorriso_handle_collision(xorriso, node, img_path, path, disk_path,
@@ -1723,7 +2102,7 @@ attach_source:;
          }
        } else {
          ret= Xorriso_restore_disk_object(xorriso, img_path, node, path,
-                                offset, bytes, (flag & (2|4|64)) | !!(flag&8));
+                     offset, bytes, (flag & (2|4|64)) | no_props | !!(flag&8));
          if(ret <= 0) {
            hret= Xorriso_eval_problem_status(xorriso, ret, 1 | 2);
            if(hret < 0)
@@ -1784,7 +2163,11 @@ int Xorriso_restore_node_array(struct XorrisO *xorriso, int flag)
                                   img_path, disk_path, hflag);
    if(ret<=0)
      goto was_problem;
- 
+
+   /* (Exclusion must already be handled when the node array gets filled,
+       because of no distinction between parameter and tree passenger here.)
+    */
+
    ret= Xorriso_handle_collision(xorriso, node, img_path, disk_path, disk_path,
                                  &stbuf_ret, 64);
    if(ret<=0 || ret==3)
@@ -1859,9 +2242,10 @@ ex:;
 int Xorriso_paste_in(struct XorrisO *xorriso, char *disk_path,
                 off_t startbyte, off_t bytecount, char *iso_rr_path, int flag)
 {
- int ret;
- char *eff_source= NULL, *eff_dest= NULL;
- struct stat stbuf;
+ int ret, no_props= 0;
+ off_t wanted_size, cap;
+ char *eff_source= NULL, *eff_dest= NULL, *reason;
+ struct stat stbuf, disk_stbuf;
  IsoNode *node;
 
  Xorriso_alloc_meM(eff_source, char, SfileadrL);
@@ -1871,18 +2255,15 @@ int Xorriso_paste_in(struct XorrisO *xorriso, char *disk_path,
                                  2|4);
  if(ret<=0)
    goto ex;
- ret= Xorriso_path_is_excluded(xorriso, disk_path, !(flag&1));
+ ret= Xorriso_path_is_excluded(xorriso, disk_path, 4 | 2 | !(flag & 1));
+ if(ret > 0 && !(flag & 1)) {
+   strcpy(xorriso->info_text,
+          "Excluded from restoring by -not_path or -not_leaf :");
+   Text_shellsafe(disk_path, xorriso->info_text, 1);
+   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "SORRY", 0);
+ }
  if(ret!=0)
    {ret= 0; goto ex;}
- ret= stat(eff_dest, &stbuf);
- if(ret!=-1 && !S_ISREG(stbuf.st_mode)) {
-   Xorriso_msgs_submit(xorriso, 0, eff_dest, 0, "ERRFILE", 0);
-   sprintf(xorriso->info_text, "-paste_in: DISK file ");
-   Text_shellsafe(eff_source, xorriso->info_text, 1); 
-   strcat(xorriso->info_text, " exists and is not a data file");
-   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, errno, "FAILURE", 0);
-   {ret= 0; goto ex;}
- }
 
  ret= Xorriso_normalize_img_path(xorriso, xorriso->wdi, iso_rr_path,
                                  eff_source, 2);
@@ -1900,9 +2281,37 @@ int Xorriso_paste_in(struct XorrisO *xorriso, char *disk_path,
    {ret= 0; goto ex;}
  }
  
- /* >>> eventually obtain parameters from file name */;
+ /* >>> ??? flag bit to obtain startbyte and bytecount from file name */;
 
- ret= Xorriso_restore(xorriso, eff_source, eff_dest, startbyte, bytecount, 8);
+ ret= lstat(eff_dest, &disk_stbuf);
+ if(ret != -1) {
+   no_props= 1 << 10;
+   wanted_size= startbyte + bytecount;
+   if(wanted_size > stbuf.st_size)
+     wanted_size= stbuf.st_size;
+   cap= wanted_size;
+   ret= Xorriso_determine_capacity(xorriso, eff_dest, &cap, &reason, 3);
+   if(ret <= 0 || (cap <= 0 && !S_ISREG(disk_stbuf.st_mode))) {
+     Xorriso_msgs_submit(xorriso, 0, eff_dest, 0, "ERRFILE", 0);
+     if(ret > 0)
+       reason= "has addressable range 0";
+     sprintf(xorriso->info_text, "-paste_in: DISK file (%s) ",
+             Ftypetxt(disk_stbuf.st_mode, 0));
+     Text_shellsafe(eff_dest, xorriso->info_text, 1); 
+     sprintf(xorriso->info_text + strlen(xorriso->info_text),
+             " exists but %s", reason);
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
+     {ret= 0; goto ex;}
+   }
+   if(cap < wanted_size && !S_ISREG(disk_stbuf.st_mode)) {
+
+     /* >>> non-regular file might be too small to take the interval */;
+
+   }
+ }
+
+ ret= Xorriso_restore(xorriso, eff_source, eff_dest, startbyte, bytecount,
+                      8 | no_props);
 ex:;
  Xorriso_free_meM(eff_source);
  Xorriso_free_meM(eff_dest);
@@ -1937,13 +2346,16 @@ int Xorriso_extract_cut(struct XorrisO *xorriso,
                                  disk_path, eff_disk_path, 2 | 4);
  if(ret<=0)
    goto ex;
- Xorriso_pacifier_reset(xorriso, 0);
- mem_lut= xorriso->last_update_time;
-
- ret= Xorriso_handle_collision(xorriso, node, img_path, eff_disk_path,
-                               disk_path, &stbuf_ret, 0);
- if(ret<=0 || ret==3)
+ ret= Xorriso_path_is_excluded(xorriso, eff_disk_path, 1 | 2 | 4);
+ if(ret < 0)
+   goto ex;
+ if(ret > 0) {
+   strcpy(xorriso->info_text,
+          "Excluded from restoring by -not_path or -not_leaf :");
+   Text_shellsafe(eff_disk_path, xorriso->info_text, 1);
+   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "SORRY", 0);
    {ret= 0; goto ex;}
+ }
 
  /* If it is a non-filtered stream from the ISO image
     and img_offset is a multiple of 2048
@@ -1957,6 +2369,13 @@ int Xorriso_extract_cut(struct XorrisO *xorriso,
    Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
    {ret= 0; goto ex;}
  }
+ ret= Xorriso_handle_collision(xorriso, node, img_path, eff_disk_path,
+                               disk_path, &stbuf_ret, 0);
+ if(ret<=0 || ret==3)
+   {ret= 0; goto ex;}
+
+ Xorriso_pacifier_reset(xorriso, 0);
+ mem_lut= xorriso->last_update_time;
  read_raw= 0;
  if((img_offset % 2048) == 0) {
    ret= Xorriso_is_plain_image_file(xorriso, node, "", 0);
@@ -1990,7 +2409,8 @@ ex:;
 }
 
 
-/* @param flag bit1= for Xorriso_check_interval(): no pacifier messages
+/* @param flag bit0= ignore node and img_path, operate on whole medium
+               bit1= for Xorriso_check_interval(): no pacifier messages
 */
 int Xorriso_read_file_data(struct XorrisO *xorriso, IsoNode *node,
                            char *img_path, char *disk_path,
@@ -1999,6 +2419,8 @@ int Xorriso_read_file_data(struct XorrisO *xorriso, IsoNode *node,
 {
  int ret, i, lba_count= 0, *start_lbas= NULL, *end_lbas= NULL, read_chunk= 16;
  int lba, count, blocks, quality, spot, bad_extract= 0;
+ int data_to_skip= 0;
+ uint32_t indev_blocks;
  off_t size= 0, file_base_bytes= 0, file_processed_bytes= 0, img_adr;
  off_t new_file_base_bytes, upto_file_bytes, start_byte= 0;
  off_t *section_sizes = NULL;
@@ -2006,25 +2428,33 @@ int Xorriso_read_file_data(struct XorrisO *xorriso, IsoNode *node,
  struct CheckmediajoB *job= NULL;
 
  upto_file_bytes= img_offset + bytes;
+ data_to_skip= img_offset % (off_t) 2048;
 
- /* >>> make Xorriso_check_interval() ready for copying in byte granularity */
- if(img_offset % (off_t) 2048) {
-   sprintf(xorriso->info_text,
-           "Image address offset is not a multiple of 2048");
-   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
-   ret= 0; goto ex;
- }
-
- ret= Xorriso__start_end_lbas(node, &lba_count, &start_lbas, &end_lbas,
-                              &section_sizes,  &size, 0);
- if(ret <= 0) {
-   Xorriso_process_msg_queues(xorriso,0);
-   sprintf(xorriso->info_text, "File object ");
-   Text_shellsafe(img_path, xorriso->info_text, 1);
-   strcat(xorriso->info_text,
-          " is currently not a data file from the loaded image");
-   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
-   goto ex;
+ if(flag & 1) {
+   lba_count= 1;
+   Xorriso_alloc_meM(start_lbas, int, 1);
+   Xorriso_alloc_meM(end_lbas, int, 1);
+   Xorriso_alloc_meM(section_sizes, off_t, 1);
+   start_lbas[0]= 0;
+   ret= Xorriso_obtain_indev_readsize(xorriso, &indev_blocks, 0);
+   if(ret > 0)
+     end_lbas[0]= indev_blocks - 1;
+   else
+     end_lbas[0]= 0x7ffffffe;
+   size= ((off_t) end_lbas[0]) * 2048;
+   section_sizes[0]= size;
+ } else {
+   ret= Xorriso__start_end_lbas(node, &lba_count, &start_lbas, &end_lbas,
+                                &section_sizes,  &size, 0);
+   if(ret <= 0) {
+     Xorriso_process_msg_queues(xorriso,0);
+     sprintf(xorriso->info_text, "File object ");
+     Text_shellsafe(img_path, xorriso->info_text, 1);
+     strcat(xorriso->info_text,
+            " is currently not a data file from the loaded image");
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
+     goto ex;
+   }
  }
  if(img_offset + bytes < size && bytes > 0)
    size= img_offset + bytes;
@@ -2067,13 +2497,14 @@ int Xorriso_read_file_data(struct XorrisO *xorriso, IsoNode *node,
      file_base_bytes= img_offset;
    }
 
-   /* Eventually omit surplus blocks */
+   /* Omit surplus blocks */
    if(new_file_base_bytes > upto_file_bytes)
      count-= (new_file_base_bytes - upto_file_bytes) / (off_t) 2048;
    /* Adjust job */
    job->data_to_offset= file_processed_bytes - img_adr + disk_offset;
    job->data_to_limit= size - file_base_bytes;
-
+   job->data_to_skip= data_to_skip;
+   data_to_skip= 0;
    file_processed_bytes+= ((off_t) count) * (off_t) 2048;
    ret= Xorriso_check_interval(xorriso, spotlist, job, lba, count, read_chunk,
                                0, (flag & 2));
@@ -2129,6 +2560,103 @@ ex:;
 }
 
 
+int Xorriso_extract_boot_images(struct XorrisO *xorriso, char *disk_dir_path,
+                                int flag)
+{
+ int ret, img_count= 0, i, was_problem= 0;
+ char **imgs= NULL, *eff_path= NULL, *cpt, *eff_namept;
+ struct stat stbuf;
+ off_t byte_offset, byte_size;
+
+ Xorriso_alloc_meM(eff_path, char, SfileadrL);
+ ret= Xorriso_normalize_img_path(xorriso, xorriso->wdx, disk_dir_path,
+                                 eff_path, 2 | 4);
+ if(ret <= 0)
+   goto ex;
+ if(strlen(eff_path) > SfileadrL - 80) {
+   sprintf(xorriso->info_text,
+           "-extract_boot_images: disk_path is too long (%lu)\n",
+           (unsigned long int) strlen(eff_path));
+   Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
+   ret= 0; goto ex;
+ }
+ ret= stat(eff_path, &stbuf);
+ if(ret == 0) {
+   if(!S_ISDIR(stbuf.st_mode)) {
+     sprintf(xorriso->info_text,
+             "-extract_boot_images: disk_path is not a directory : ");
+     Text_shellsafe(eff_path, xorriso->info_text, 1);
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, 0, "FAILURE", 0);
+     ret= 0; goto ex;
+   }
+ } else {
+   ret= mkdir(eff_path, 0777);
+   if(ret == -1) {
+     sprintf(xorriso->info_text,
+             "-extract_boot_images: cannot create directory : ");
+     Text_shellsafe(eff_path, xorriso->info_text, 1);
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, errno, "FAILURE", 0);
+     ret= 0; goto ex;
+   }
+ }
+ strcat(eff_path, "/");
+ eff_namept= eff_path + strlen(eff_path);
+
+ ret= Xorriso_list_boot_images(xorriso, &imgs, &img_count, 0);
+ if(ret <= 0)
+   goto ex;
+
+ /* Interpret list and create files */
+ for(i= 0; i < img_count; i++) {
+   ret= Xorriso_eval_problem_status(xorriso, 1, 1 | 2);
+   if(ret < 0)
+     {ret= 0; goto ex;}
+   cpt= strchr(imgs[i], '/');
+   if(cpt == NULL)
+ continue;
+   *cpt= 0;
+   cpt+= 2;
+   ret= Sfile_text_to_off_t(cpt, &byte_offset, 0);
+   if(ret <= 0)
+ continue;
+   cpt+= ret;
+   if(*cpt == 0)
+ continue;
+   cpt++;
+   ret= Sfile_text_to_off_t(cpt, &byte_size, 0);
+   if(ret <= 0)
+ continue;
+
+   strcpy(eff_namept, imgs[i]);
+   sprintf(xorriso->info_text, "%s : offset=%.f size=%.f\n",
+           eff_path, (double) byte_offset, (double) byte_size);
+   Xorriso_info(xorriso, 0);
+
+   ret= stat(eff_path, &stbuf);
+   if(ret != -1) {
+     sprintf(xorriso->info_text,
+             "-extract_boot_images: File already exists on disk: ");
+     Text_shellsafe(eff_path, xorriso->info_text, 1);
+     Xorriso_msgs_submit(xorriso, 0, xorriso->info_text, errno, "FAILURE", 0);
+ continue;
+   }
+   ret= Xorriso_read_file_data(xorriso, NULL, NULL, eff_path,
+                               byte_offset, (off_t) 0, byte_size, 1);
+   if(ret <= 0)
+     was_problem= 1;
+ }
+ ret= Xorriso_eval_problem_status(xorriso, 1, 1 | 2);
+ if(ret < 0 || was_problem)
+   {ret= 0; goto ex;}
+ 
+ ret= 1;
+ex:;
+ Xorriso_free_meM(eff_path);
+ Xorriso_list_boot_images(xorriso, &imgs, &img_count, 1 << 15);
+ return(ret);
+}
+
+
 /* @param node      Opaque handle to IsoNode which is to be inquired instead of                     path if it is not NULL.
    @param path      is used as address if node is NULL.
    @param flag      bit0= do not report to result but only indicate outcome
