@@ -2,7 +2,7 @@
 /* xorriso - libisoburn higher level API which creates, loads, manipulates
              and burns ISO 9660 filesystem images.
 
-   Copyright 2007-2021 Thomas Schmitt, <scdbackup@gmx.net>
+   Copyright 2007-2023 Thomas Schmitt, <scdbackup@gmx.net>
 
    Provided under GPL version 2 or later.
 
@@ -74,7 +74,7 @@ struct XorrisO;
 */
 #define Xorriso_header_version_majoR  1
 #define Xorriso_header_version_minoR  5
-#define Xorriso_header_version_micrO  4
+#define Xorriso_header_version_micrO  6
 
 
 /** If needed: Something like ".pl01" to indicate a bug fix. Normally empty.
@@ -867,6 +867,9 @@ int Xorriso_sieve_dispose(struct XorrisO *xorriso, int flag);
                      argv[3]= Volume Id
     "Image size   :" up to 1 result from -print_size
                      argv[0]= number of blocks with "s" appended
+    "Indev feature:" up to 10000 results from -assess_indev_features plain
+                     argv[0]= name
+                     argv[1]= value
     "Jigdo files  :" up to 1 result from -list_extras
                      argv[0]= "yes" or "no"
     "Local ACL    :" up to 1 result from -list_extras
@@ -908,6 +911,9 @@ int Xorriso_sieve_dispose(struct XorrisO *xorriso, int flag);
     "Modif. Time  :" up to 1 result from -pvd_info
                      argv[0]= YYYYMMDDhhmmsscc
                      (Note: prefix is "Modif. Time  : ")
+    "Overburnt by :" up to 2 results from -toc
+                     argv[0]= number of readable blocks beyond ATIP leadout
+                     (Note: only reported with overburnt CD-RW)
     "PVD address  :" up to 1 result from -pvd_info
                      argv[0]= block address with "s" appended
     "Preparer Id  :" up to 1 result from -pvd_info
@@ -1290,6 +1296,11 @@ int Xorriso_option_as(struct XorrisO *xorriso, int argc, char **argv,
 int Xorriso_option_assert_volid(struct XorrisO *xorriso, char *pattern,
                                 char *severity, int flag);
 
+/* Command -assess_indev_features */
+/* @since 1.5.6 */
+int Xorriso_option_assess_indev_features(struct XorrisO *xorriso, char *mode,
+                                         int flag);
+
 /* Command -auto_charset "on"|"off" */
 /* @since 0.3.8 */
 int Xorriso_option_auto_charset(struct XorrisO *xorriso, char *mode, int flag);
